@@ -1,11 +1,15 @@
-/*! Bit Management
+/*! Bit management
 
-The `Bits` trait defines constants and free functions suitable for managing bit
-storage of a fundamental, and is the constraint for the storage type of the data
-structures of the rest of the crate.
+The `Bits` trait defines constants and associated functions suitable for
+managing the bit patterns of a fundamental, and is the constraint for the
+storage type of the data structures of the rest of the crate.
+
+The other types in this module provide stronger rules about how indices map to
+concrete bits in fundamental elements. They are implementation details, and are
+not exported in the prelude.
 !*/
 
-use crate::Cursor;
+use crate::cursor::Cursor;
 use core::{
 	cmp::Eq,
 	convert::From,
@@ -37,6 +41,12 @@ use core::{
 	},
 };
 
+#[cfg(feature = "atomic")]
+use crate::atomic::Atomic;
+
+#[cfg(feature = "atomic")]
+use core::sync::atomic;
+
 /** Generalizes over the fundamental types for use in `bitvec` data structures.
 
 This trait must only be implemented on unsigned integer primitives with full
@@ -78,29 +88,66 @@ pub trait Bits:
 	+ Sync
 	+ UpperHex
 {
-	/// The size, in bits, of this type.
-	const SIZE: u8 = size_of::<Self>() as u8 * 8;
+	/// The width, in bits, of this type.
+	const BITS: u8 = size_of::<Self>() as u8 * 8;
 
-	/// The number of bits required to index the type. This is always
-	/// log<sub>2</sub> of the type’s bit size.
-	///
-	/// Incidentally, this can be computed as `Self::SIZE.trailing_zeros()` once
-	/// that becomes a valid constexpr.
-	const BITS: u8; // = Self::SIZE.trailing_zeros() as u8;
+	/// The number of bits required to index a bit inside the type. This is
+	/// always log<sub>2</sub> of the type’s bit width.
+	const INDX: u8 = Self::BITS.trailing_zeros() as u8;
 
-	/// The bitmask to turn an arbitrary `usize` into a bit index. Bit indices
+	/// The bitmask to turn an arbitrary number into a bit index. Bit indices
 	/// are always stored in the lowest bits of an index value.
-	const MASK: u8 = Self::SIZE - 1;
+	const MASK: u8 = Self::BITS - 1;
 
-	/// Name of the implementing type.
+	/// Name of the implementing type. This is only necessary until the compiler
+	/// stabilizes `type_name()`.
 	const TYPENAME: &'static str;
 
+	/// Atomic version of the storage type, to have properly fenced access.
+	#[cfg(feature = "atomic")]
+	#[doc(hidden)]
+	type Atom: Atomic<Self>;
+
+	/// Performs a synchronized load on the underlying element.
+	///
+	/// # Parameters
+	///
+	/// - `&self`
+	///
+	/// # Returns
+	///
+	/// The element referred to by the `self` reference, loaded synchronously
+	/// after any in-progress accesses have concluded.
+	#[cfg(feature = "atomic")]
+	#[inline(always)]
+	fn load(&self) -> Self {
+		let aptr = self as *const Self as *const Self::Atom;
+		unsafe { &*aptr }.get()
+	}
+
+	/// Performs an unsynchronized load on the underlying element.
+	///
+	/// As atomic operations are unavailable, this is a standard dereference.
+	///
+	/// # Parameters
+	///
+	/// - `&self`
+	///
+	/// # Returns
+	///
+	/// The referent element.
+	#[cfg(not(feature = "atomic"))]
+	#[inline(always)]
+	fn load(&self) -> Self {
+		*self
+	}
+
 	/// Sets a specific bit in an element to a given value.
 	///
 	/// # Parameters
 	///
-	/// - `place`: A bit index in the element, from `0` at `LSb` to `Self::MASK`
-	///   at `MSb`. The bit under this index will be set according to `value`.
+	/// - `place`: A bit index in the element, from `0` to `Self::MASK`. The bit
+	///   under this index will be set according to `value`.
 	/// - `value`: A Boolean value, which sets the bit on `true` and unsets it
 	///   on `false`.
 	///
@@ -111,51 +158,106 @@ pub trait Bits:
 	///
 	/// # Panics
 	///
-	/// This function panics if `place` is not less than `T::SIZE`, in order
-	/// to avoid index out of range errors.
+	/// This function panics if `place` is not less than `T::BITS`, in order to
+	/// avoid index out of range errors.
 	///
 	/// # Examples
 	///
 	/// This example sets and unsets bits in a byte.
 	///
 	/// ```rust
-	/// use bitvec::{Bits, LittleEndian};
-	/// let mut elt: u8 = 0;
-	/// elt.set::<LittleEndian>(0.into(), true);
-	/// assert_eq!(elt, 0b0000_0001);
-	/// elt.set::<LittleEndian>(4.into(), true);
-	/// assert_eq!(elt, 0b0001_0001);
-	/// elt.set::<LittleEndian>(0.into(), false);
-	/// assert_eq!(elt, 0b0001_0000);
+	/// use bitvec::prelude::{
+	///   Bits,
+	///   BigEndian,
+	///   LittleEndian,
+	/// };
+	///
+	/// let mut elt: u16 = 0;
+	///
+	/// elt.set::<BigEndian>(1.into(), true);
+	/// assert_eq!(elt, 0b0100_0000__0000_0000);
+	/// elt.set::<LittleEndian>(1.into(), true);
+	/// assert_eq!(elt, 0b0100_0000__0000_0010);
+	///
+	/// elt.set::<BigEndian>(1.into(), false);
+	/// assert_eq!(elt, 0b0000_0000__0000_0010);
+	/// elt.set::<LittleEndian>(1.into(), false);
+	/// assert_eq!(elt, 0);
 	/// ```
 	///
 	/// This example overruns the index, and panics.
 	///
 	/// ```rust,should_panic
-	/// use bitvec::{Bits, LittleEndian};
+	/// use bitvec::prelude::{Bits, BigEndian};
 	/// let mut elt: u8 = 0;
-	/// elt.set::<LittleEndian>(8.into(), true);
+	/// elt.set::<BigEndian>(8.into(), true);
 	/// ```
-	fn set<C: Cursor>(&mut self, place: BitIdx, value: bool) {
-		let place: BitPos = C::at::<Self>(place);
-		assert!(
-			*place < Self::SIZE,
-			"Index out of range: {} overflows {}",
-			*place,
-			Self::SIZE,
-		);
-		//  Blank the selected bit
-		*self &= !(Self::from(1u8) << *place);
-		//  Set the selected bit
-		*self |= Self::from(value as u8) << *place;
+	#[inline(always)]
+	fn set<C>(&mut self, place: BitIdx, value: bool)
+	where C: Cursor {
+		self.set_at(C::at::<Self>(place), value)
+	}
+
+	/// Sets a specific bit in an element to a given value.
+	///
+	/// # Parameters
+	///
+	/// - `place`: A bit *position* in the element, where `0` is the LSbit and
+	///   `Self::MASK` is the MSbit.
+	/// - `value`: A Boolean value, which sets the bit high on `true` and unsets
+	///   it low on `false`.
+	///
+	/// # Panics
+	///
+	/// This function panics if `place` is not less than `T::BITS`, in order to
+	/// avoid index out of range errors.
+	///
+	/// # Examples
+	///
+	/// This example sets and unsets bits in a byte.
+	///
+	/// ```rust
+	/// use bitvec::prelude::Bits;
+	/// let mut elt: u8 = 0;
+	/// elt.set_at(0.into(), true);
+	/// assert_eq!(elt, 0b0000_0001);
+	/// elt.set_at(7.into(), true);
+	/// assert_eq!(elt, 0b1000_0001);
+	/// ```
+	///
+	/// This example overshoots the width, and panics.
+	///
+	/// ```rust,should_panic
+	/// use bitvec::prelude::Bits;
+	/// let mut elt: u8 = 0;
+	/// elt.set_at(8.into(), true);
+	/// ```
+	fn set_at(&mut self, place: BitPos, value: bool) {
+		#[cfg(feature = "atomic")] {
+			let aptr = self as *const Self as *const Self::Atom;
+			if value {
+				unsafe { &*aptr }.set(place);
+			}
+			else {
+				unsafe { &*aptr }.clear(place);
+			}
+		}
+		#[cfg(not(feature = "atomic"))] {
+			if value {
+				*self |= Self::mask_at(place);
+			}
+			else {
+				*self &= !Self::mask_at(place);
+			}
+		}
 	}
 
 	/// Gets a specific bit in an element.
 	///
 	/// # Parameters
 	///
-	/// - `place`: A bit index in the element, from `0` at `LSb` to `Self::MASK`
-	///   at `MSb`. The bit under this index will be retrieved as a `bool`.
+	/// - `place`: A bit index in the element, from `0` to `Self::MASK`. The bit
+	///   under this index will be retrieved as a `bool`.
 	///
 	/// # Returns
 	///
@@ -168,37 +270,158 @@ pub trait Bits:
 	///
 	/// # Panics
 	///
-	/// This function panics if `place` is not less than `T::SIZE`, in order
-	/// to avoid index out of range errors.
+	/// This function panics if `place` is not less than `T::BITS`, in order to
+	/// avoid index out of range errors.
 	///
 	/// # Examples
 	///
 	/// This example gets two bits from a byte.
 	///
 	/// ```rust
-	/// use bitvec::{Bits, LittleEndian};
-	/// let elt: u8 = 0b0000_0100;
-	/// assert!(!elt.get::<LittleEndian>(1.into()));
-	/// assert!(elt.get::<LittleEndian>(2.into()));
-	/// assert!(!elt.get::<LittleEndian>(3.into()));
+	/// use bitvec::prelude::{Bits, BigEndian};
+	/// let elt: u8 = 0b0010_0000;
+	/// assert!(!elt.get::<BigEndian>(1.into()));
+	/// assert!(elt.get::<BigEndian>(2.into()));
+	/// assert!(!elt.get::<BigEndian>(3.into()));
 	/// ```
 	///
 	/// This example overruns the index, and panics.
 	///
 	/// ```rust,should_panic
-	/// use bitvec::{Bits, LittleEndian};
-	/// 0u8.get::<LittleEndian>(8.into());
+	/// use bitvec::prelude::{Bits, BigEndian};
+	/// 0u8.get::<BigEndian>(8.into());
 	/// ```
-	fn get<C: Cursor>(&self, place: BitIdx) -> bool {
-		let place: BitPos = C::at::<Self>(place);
+	fn get<C>(&self, place: BitIdx) -> bool
+	where C: Cursor {
+		self.get_at(C::at::<Self>(place))
+	}
+
+	/// Gets a specific bit in an element.
+	///
+	/// # Parameters
+	///
+	/// - `place`: A bit *position* in the element, from `0` at LSbit to
+	///   `Self::MASK` at MSbit. The bit under this position will be retrieved
+	///   as a `bool`.
+	///
+	/// # Returns
+	///
+	/// The value of the bit under `place`, as a `bool`.
+	///
+	/// # Panics
+	///
+	/// This function panics if `place` is not less than `T::BITS`, in order to
+	/// avoid index out of range errors.
+	///
+	/// # Examples
+	///
+	/// This example gets two bits from a byte.
+	///
+	/// ```rust
+	/// use bitvec::prelude::Bits;
+	/// let elt: u8 = 0b0010_0000;
+	/// assert!(!elt.get_at(4.into()));
+	/// assert!(elt.get_at(5.into()));
+	/// assert!(!elt.get_at(6.into()));
+	/// ```
+	///
+	/// This example overruns the index, and panics.
+	///
+	/// ```rust,should_panic
+	/// use bitvec::prelude::Bits;
+	/// 0u8.get_at(8.into());
+	/// ```
+	fn get_at(&self, place: BitPos) -> bool {
+		self.load() & Self::mask_at(place) != Self::from(0u8)
+	}
+
+	/// Produces the bit mask which selects only the bit at the requested
+	/// position.
+	///
+	/// This mask must be inverted in order to clear the bit.
+	///
+	/// # Parameters
+	///
+	/// - `place`: The bit position for which to create a bitmask.
+	///
+	/// # Returns
+	///
+	/// The one-hot encoding of the bit position index.
+	///
+	/// # Panics
+	///
+	/// This function panics if `place` is not less than `T::BITS`, in order to
+	/// avoid index out of range errors.
+	///
+	/// # Examples
+	///
+	/// This example produces the one-hot encodings for indices.
+	///
+	/// ```rust
+	/// use bitvec::prelude::Bits;
+	///
+	/// assert_eq!(u8::mask_at(0.into()), 0b0000_0001);
+	/// assert_eq!(u8::mask_at(1.into()), 0b0000_0010);
+	/// assert_eq!(u8::mask_at(2.into()), 0b0000_0100);
+	/// assert_eq!(u8::mask_at(3.into()), 0b0000_1000);
+	/// assert_eq!(u8::mask_at(4.into()), 0b0001_0000);
+	/// assert_eq!(u8::mask_at(5.into()), 0b0010_0000);
+	/// assert_eq!(u8::mask_at(6.into()), 0b0100_0000);
+	/// assert_eq!(u8::mask_at(7.into()), 0b1000_0000);
+	///
+	/// assert_eq!(u16::mask_at(8.into()),  0b0000_0001__0000_0000);
+	/// assert_eq!(u16::mask_at(9.into()),  0b0000_0010__0000_0000);
+	/// assert_eq!(u16::mask_at(10.into()), 0b0000_0100__0000_0000);
+	/// assert_eq!(u16::mask_at(11.into()), 0b0000_1000__0000_0000);
+	/// assert_eq!(u16::mask_at(12.into()), 0b0001_0000__0000_0000);
+	/// assert_eq!(u16::mask_at(13.into()), 0b0010_0000__0000_0000);
+	/// assert_eq!(u16::mask_at(14.into()), 0b0100_0000__0000_0000);
+	/// assert_eq!(u16::mask_at(15.into()), 0b1000_0000__0000_0000);
+	///
+	/// assert_eq!(u32::mask_at(16.into()), 1 << 16);
+	/// assert_eq!(u32::mask_at(24.into()), 1 << 24);
+	/// assert_eq!(u32::mask_at(31.into()), 1 << 31);
+	///
+	/// # #[cfg(target_pointer_width = "64")] {
+	/// assert_eq!(u64::mask_at(32.into()), 1 << 32);
+	/// assert_eq!(u64::mask_at(48.into()), 1 << 48);
+	/// assert_eq!(u64::mask_at(63.into()), 1 << 63);
+	/// # }
+	/// ```
+	///
+	/// These examples ensure that indices panic when out of bounds.
+	///
+	/// ```rust,should_panic
+	/// use bitvec::prelude::Bits;
+	/// u8::mask_at(8.into());
+	/// ```
+	///
+	/// ```rust,should_panic
+	/// use bitvec::prelude::Bits;
+	/// u16::mask_at(16.into());
+	/// ```
+	///
+	/// ```rust,should_panic
+	/// use bitvec::prelude::Bits;
+	/// u32::mask_at(32.into());
+	/// ```
+	///
+	/// ```rust,should_panic
+	/// # #[cfg(target_pointer_width = "64")] {
+	/// use bitvec::prelude::Bits;
+	/// u64::mask_at(64.into());
+	/// # }
+	/// ```
+	#[inline(always)]
+	fn mask_at(place: BitPos) -> Self {
 		assert!(
-			*place < Self::SIZE,
+			*place < Self::BITS,
 			"Index out of range: {} overflows {}",
 			*place,
-			Self::SIZE,
+			Self::BITS,
 		);
-		//  Shift down so the targeted bit is in LSb, then blank all other bits.
-		(*self >> *place) & Self::from(1) == Self::from(1)
+		//  Pad 1 to the correct width, then shift up to the correct bit place.
+		Self::from(1u8) << *place
 	}
 
 	/// Counts how many bits in `self` are set to `1`.
@@ -218,7 +441,7 @@ pub trait Bits:
 	/// # Examples
 	///
 	/// ```rust
-	/// use bitvec::Bits;
+	/// use bitvec::prelude::Bits;
 	/// assert_eq!(Bits::count_ones(&0u8), 0);
 	/// assert_eq!(Bits::count_ones(&128u8), 1);
 	/// assert_eq!(Bits::count_ones(&192u8), 2);
@@ -233,7 +456,7 @@ pub trait Bits:
 	/// [`u64::count_ones`]: https://doc.rust-lang.org/stable/std/primitive.u64.html#method.count_ones
 	#[inline(always)]
 	fn count_ones(&self) -> usize {
-		u64::count_ones((*self).into()) as usize
+		u64::count_ones((self.load()).into()) as usize
 	}
 
 	/// Counts how many bits in `self` are set to `0`.
@@ -254,7 +477,7 @@ pub trait Bits:
 	/// # Examples
 	///
 	/// ```rust
-	/// use bitvec::Bits;
+	/// use bitvec::prelude::Bits;
 	/// assert_eq!(Bits::count_zeros(&0u8), 8);
 	/// assert_eq!(Bits::count_zeros(&1u8), 7);
 	/// assert_eq!(Bits::count_zeros(&3u8), 6);
@@ -269,7 +492,23 @@ pub trait Bits:
 	/// [`u64::count_ones`]: https://doc.rust-lang.org/stable/std/primitive.u64.html#method.count_ones
 	#[inline(always)]
 	fn count_zeros(&self) -> usize {
-		u64::count_ones((!*self).into()) as usize
+		//  invert (0 becomes 1, 1 becomes 0), zero-extend, count ones
+		u64::count_ones((!self.load()).into()) as usize
+	}
+
+	/// Extends a single bit to fill the entire element.
+	///
+	/// # Parameters
+	///
+	/// - `bit`: The bit to extend.
+	///
+	/// # Returns
+	///
+	/// An element with all bits set to the input.
+	#[inline]
+	fn bits(bit: bool) -> Self {
+		//  convert 0 to !0 and 1 to 0, then invert.
+		!Self::from((bit as u8).wrapping_sub(1))
 	}
 }
 
@@ -283,14 +522,17 @@ ordering. Values of `BitIdx` refer strictly to abstract ordering, and not to the
 actual position in an element, so `BitIdx(0)` is the first bit in an element,
 but is not required to be the electrical `LSb`, `MSb`, or any other.
 
-[`Cursor`]: ../trait.Cursor.html
+[`Cursor`]: ../cursor/trait.Cursor.html
 **/
 #[derive(Clone, Copy, Debug, Default, Eq, Hash, Ord, PartialEq, PartialOrd)]
+#[doc(hidden)]
 pub struct BitIdx(pub(crate) u8);
 
 impl BitIdx {
 	/// Checks if the index is valid for a type.
 	///
+	/// Indices are valid in the range `0 .. T::BITS`.
+	///
 	/// # Parameters
 	///
 	/// - `self`: The index to validate.
@@ -302,8 +544,31 @@ impl BitIdx {
 	/// # Type Parameters
 	///
 	/// - `T: Bits`: The storage type used to determine index validity.
-	pub fn is_valid<T: Bits>(self) -> bool {
-		*self < T::SIZE
+	#[inline]
+	pub fn is_valid<T>(self) -> bool
+	where T: Bits {
+		*self < T::BITS
+	}
+
+	/// Checks if the index is valid as a tail index for a type.
+	///
+	/// Tail indices are vaild in the range `1 ..= T::BITS`.
+	///
+	/// # Parameters
+	///
+	/// - `self`: The index to validate as a tail.
+	///
+	/// # Returns
+	///
+	/// Whether the index is valid as a tail for the storage type in question.
+	///
+	/// # Type Parameters
+	///
+	/// - `T: Bits`: The storage used to determine index tail validity.
+	#[inline]
+	pub fn is_valid_tail<T>(self) -> bool
+	where T: Bits {
+		*self > 0 && *self <= T::BITS
 	}
 
 	/// Increments a cursor to the next value, wrapping if needed.
@@ -324,7 +589,7 @@ impl BitIdx {
 	///
 	/// # Panics
 	///
-	/// This method panics if `self` is not less than `T::SIZE`, in order to
+	/// This method panics if `self` is not less than `T::BITS`, in order to
 	/// avoid index out of range errors.
 	///
 	/// # Examples
@@ -334,7 +599,7 @@ impl BitIdx {
 	/// ```rust
 	/// # #[cfg(feature = "testing")] {
 	/// use bitvec::testing::BitIdx;
-	/// assert_eq!(BitIdx::from(3).incr::<u8>(), (4.into(), false));
+	/// assert_eq!(BitIdx::from(6).incr::<u8>(), (7.into(), false));
 	/// # }
 	/// ```
 	///
@@ -346,18 +611,93 @@ impl BitIdx {
 	/// assert_eq!(BitIdx::from(7).incr::<u8>(), (0.into(), true));
 	/// # }
 	/// ```
-	pub fn incr<T: Bits>(self) -> (Self, bool) {
+	pub fn incr<T>(self) -> (Self, bool)
+	where T: Bits {
+		let val = *self;
 		assert!(
-			*self < T::SIZE,
+			self.is_valid::<T>(),
 			"Index out of range: {} overflows {}",
-			*self,
-			T::SIZE,
+			val,
+			T::BITS,
 		);
-		let next = (*self).wrapping_add(1) & T::MASK;
+		let next = val.wrapping_add(1) & T::MASK;
 		(next.into(), next == 0)
 	}
 
-	/// Decrements a cursor to the previous value, wrapping if needed.
+	/// Increments a tail cursor to the next value, wrapping if needed.
+	///
+	/// Tail cursors have the domain `1 ..= T::BITS`, with the exception that
+	/// the tail of an empty domain is `0`. As such, it is valid for a tail to
+	/// increment *from* `0`, but will never return to it.
+	///
+	/// # Parameters
+	///
+	/// - `self`: The original tail cursor.
+	///
+	/// # Returns
+	///
+	/// - `Self`: An incremented tail cursor.
+	/// - `bool`: Marks whether the increment crossed an element boundary
+	///   (including from `0` to `1`).
+	///
+	/// # Type Parameters
+	///
+	/// - `T: Bits`: The storage type for which the increment will be
+	///   calculated.
+	///
+	/// # Panics
+	///
+	/// This method panics if `self` is outside the range `0 ..= T::BITS`, in
+	/// order to avoid index out of range errors.
+	///
+	/// # Examples
+	///
+	/// This example increments from zero.
+	///
+	/// ```rust
+	/// # #[cfg(feature = "testing")] {
+	/// use bitvec::testing::BitIdx;
+	/// assert_eq!(BitIdx::from(0).incr_tail::<u8>(), (1.into(), true));
+	/// # }
+	/// ```
+	///
+	/// This example increments inside an element.
+	///
+	/// ```rust
+	/// # #[cfg(feature = "testing")] {
+	/// use bitvec::testing::BitIdx;
+	/// assert_eq!(BitIdx::from(7).incr_tail::<u8>(), (8.into(), false));
+	/// # }
+	/// ```
+	///
+	/// This example increments at the high edge, and wraps to the next element.
+	///
+	/// ```rust
+	/// # #[cfg(feature = "testing")] {
+	/// use bitvec::testing::BitIdx;
+	/// assert_eq!(BitIdx::from(8).incr_tail::<u8>(), (1.into(), true));
+	/// # }
+	/// ```
+	pub fn incr_tail<T>(self) -> (Self, bool)
+	where T: Bits {
+		let val = *self;
+		//  Permit 0 ..= T::BITS, rather than 1 ..= T::BITS, for the empty tail.
+		assert!(
+			val <= T::BITS,
+			"Index out of range: {} exceeds {}",
+			val,
+			T::BITS,
+		);
+		if val == T::BITS {
+			(1.into(), true)
+		}
+		else {
+			//  Signal wrap if the tail was empty
+			(val.wrapping_add(1).into(), val == 0)
+		}
+	}
+
+	/// Decrements a cursor to the prior value, wrapping if needed.
 	///
 	/// # Parameters
 	///
@@ -375,7 +715,7 @@ impl BitIdx {
 	///
 	/// # Panics
 	///
-	/// This method panics if `self` is not less than `T::SIZE`, in order to
+	/// This method panics if `self` is not less than `T::BITS`, in order to
 	/// avoid index out of range errors.
 	///
 	/// # Examples
@@ -385,29 +725,104 @@ impl BitIdx {
 	/// ```rust
 	/// # #[cfg(feature = "testing")] {
 	/// use bitvec::testing::BitIdx;
-	/// assert_eq!(BitIdx::from(5).decr::<u8>(), (4.into(), false));
+	/// assert_eq!(BitIdx::from(1).decr::<u8>(), (0.into(), false));
 	/// # }
 	/// ```
 	///
-	/// This example decrements at the low edge, and wraps to the previous
-	/// element.
+	/// This example decrements at the low edge, and wraps to the prior element.
 	///
 	/// ```rust
 	/// # #[cfg(feature = "testing")] {
 	/// use bitvec::testing::BitIdx;
 	/// assert_eq!(BitIdx::from(0).decr::<u8>(), (7.into(), true));
 	/// # }
-	pub fn decr<T: Bits>(self) -> (Self, bool) {
+	pub fn decr<T>(self) -> (Self, bool)
+	where T: Bits {
+		let val = *self;
 		assert!(
-			*self < T::SIZE,
+			self.is_valid::<T>(),
 			"Index out of range: {} overflows {}",
-			*self,
-			T::SIZE,
+			val,
+			T::BITS,
 		);
-		let (prev, wrap) = (*self).overflowing_sub(1);
+		let (prev, wrap) = val.overflowing_sub(1);
 		((prev & T::MASK).into(), wrap)
 	}
 
+	/// Decrements a tail cursor to the prior value, wrapping if needed.
+	///
+	/// Tail cursors have the domain `1 ..= T::BITS`. It is forbidden to
+	/// decrement the tail of an empty slice, so this method disallows tails of
+	/// value zero.
+	///
+	/// # Parameters
+	///
+	/// - `self`: The original tail cursor.
+	///
+	/// # Returns
+	///
+	/// - `Self`: A decremented tail cursor.
+	/// - `bool`: Marks whether the decrement crossed an element boundary (from
+	///   `1` to `T::BITS`).
+	///
+	/// # Type Parameters
+	///
+	/// - `T: Bits`: The storage type for which the decrement will be
+	///   calculated.
+	///
+	/// # Panics
+	///
+	/// This method panics if `self` is outside the range `1 ..= T::BITS`, in
+	/// order to avoid index out of range errors.
+	///
+	/// # Examples
+	///
+	/// This example demonstrates that the zero tail cannot decrement.
+	///
+	/// ```rust,should_panic
+	/// # #[cfg(feature = "testing")] {
+	/// use bitvec::testing::BitIdx;
+	/// BitIdx::from(0).decr_tail::<u8>();
+	/// # }
+	/// # #[cfg(not(feature = "testing"))]
+	/// # panic!("Keeping the test green even when this can't run");
+	/// ```
+	///
+	/// This example decrements inside an element.
+	///
+	/// ```rust
+	/// # #[cfg(feature = "testing")] {
+	/// use bitvec::testing::BitIdx;
+	/// assert_eq!(BitIdx::from(2).decr_tail::<u8>(), (1.into(), false));
+	/// # }
+	/// ```
+	///
+	/// This example decrements at the low edge, and wraps to the prior element.
+	///
+	/// ```rust
+	/// # #[cfg(feature = "testing")] {
+	/// use bitvec::testing::BitIdx;
+	/// assert_eq!(BitIdx::from(1).decr_tail::<u8>(), (8.into(), true));
+	/// # }
+	/// ```
+	pub fn decr_tail<T>(self) -> (Self, bool)
+	where T: Bits {
+		let val = *self;
+		//  The empty tail cannot decrement.
+		assert!(
+			self.is_valid_tail::<T>(),
+			"Index out of range: {} departs 1 ..= {}",
+			val,
+			T::BITS,
+		);
+		if val == 1 {
+			(T::BITS.into(), true)
+		}
+		else {
+			(val.wrapping_sub(1).into(), false)
+		}
+	}
+
 	/// Finds the destination bit a certain distance away from a starting bit.
 	///
 	/// This produces the number of elements to move, and then the bit index of
@@ -416,7 +831,7 @@ impl BitIdx {
 	/// # Parameters
 	///
 	/// - `self`: The bit index in an element of the starting position. This
-	///   must be in the domain `0 .. T::SIZE`.
+	///   must be in the domain `0 .. T::BITS`.
 	/// - `by`: The number of bits by which to move. Negative values move
 	///   downwards in memory: towards `LSb`, then starting again at `MSb` of
 	///   the prior element in memory (decreasing address). Positive values move
@@ -428,7 +843,7 @@ impl BitIdx {
 	/// - `isize`: The number of elements by which to change the caller’s
 	///   element cursor. This value can be passed directly into [`ptr::offset`]
 	/// - `BitIdx`: The bit index of the destination bit in the newly selected
-	///   element. This will always be in the domain `0 .. T::SIZE`. This
+	///   element. This will always be in the domain `0 .. T::BITS`. This
 	///   value can be passed directly into [`Cursor`] functions to compute the
 	///   correct place in the element.
 	///
@@ -438,7 +853,7 @@ impl BitIdx {
 	///
 	/// # Panics
 	///
-	/// This function panics if `from` is not less than `T::SIZE`, in order
+	/// This function panics if `from` is not less than `T::BITS`, in order
 	/// to avoid index out of range errors.
 	///
 	/// # Safety
@@ -474,30 +889,32 @@ impl BitIdx {
 	/// # }
 	/// ```
 	///
-	/// [`Cursor`]: ../trait.Cursor.html
+	/// [`Cursor`]: ../cursor/trait.Cursor.html
 	/// [`ptr::offset`]: https://doc.rust-lang.org/stable/std/primitive.pointer.html#method.offset
-	pub fn offset<T: Bits>(self, by: isize) -> (isize, Self) {
+	pub fn offset<T>(self, by: isize) -> (isize, Self)
+	where T: Bits {
+		let val = *self;
 		assert!(
-			*self < T::SIZE,
+			val < T::BITS,
 			"Index out of range: {} overflows {}",
-			*self,
-			T::SIZE,
+			val,
+			T::BITS,
 		);
 		//  If the `isize` addition does not overflow, then the sum can be used
 		//  directly.
-		if let (far, false) = by.overflowing_add(*self as isize) {
-			//  If `far` is in the domain `0 .. T::SIZE`, then the offset did
+		if let (far, false) = by.overflowing_add(val as isize) {
+			//  If `far` is in the domain `0 .. T::BITS`, then the offset did
 			//  not depart the element.
-			if far >= 0 && far < T::SIZE as isize {
+			if far >= 0 && far < T::BITS as isize {
 				(0, (far as u8).into())
 			}
 			//  If `far` is negative, then the offset leaves the initial element
-			//  going down. If `far` is not less than `T::SIZE`, then the
+			//  going down. If `far` is not less than `T::BITS`, then the
 			//  offset leaves the initial element going up.
 			else {
 				//  `Shr` on `isize` sign-extends
 				(
-					far >> T::BITS,
+					far >> T::INDX,
 					((far & (T::MASK as isize)) as u8).into(),
 				)
 			}
@@ -505,15 +922,15 @@ impl BitIdx {
 		//  If the `isize` addition overflows, then the `by` offset is positive.
 		//  Add as `usize` and use that. This is guaranteed not to overflow,
 		//  because `isize -> usize` doubles the domain, but `self` is limited
-		//  to `0 .. T::SIZE`.
+		//  to `0 .. T::BITS`.
 		else {
-			let far = *self as usize + by as usize;
+			let far = val as usize + by as usize;
 			//  This addition will always result in a `usize` whose lowest
-			//  `T::BITS` bits are the bit index in the destination element,
+			//  `T::INDX` bits are the bit index in the destination element,
 			//  and the rest of the high bits (shifted down) are the number of
 			//  elements by which to advance.
 			(
-				(far >> T::BITS) as isize,
+				(far >> T::INDX) as isize,
 				((far & (T::MASK as usize)) as u8).into(),
 			)
 		}
@@ -531,7 +948,7 @@ impl BitIdx {
 	/// - `usize`: The number of elements `T` included in the span. This will
 	///   be in the domain `1 .. usize::max_value()`.
 	/// - `BitIdx`: The index of the first bit *after* the span. This will be in
-	///   the domain `1 ..= T::SIZE`.
+	///   the domain `1 ..= T::BITS`.
 	///
 	/// # Type Parameters
 	///
@@ -547,32 +964,31 @@ impl BitIdx {
 	/// assert_eq!(BitIdx::from(0).span::<u8>(8), (1, 8.into()))
 	/// # }
 	/// ```
-	pub fn span<T: Bits>(self, len: usize) -> (usize, BitIdx) {
+	pub fn span<T>(self, len: usize) -> (usize, BitIdx)
+	where T: Bits {
 		//  Number of bits in the head *element*. Domain 32 .. 0.
-		let bits_in_head = (T::SIZE - *self) as usize;
-		//  If there are n bits live between the head cursor (which marks the
+		let bits_in_head = (T::BITS - *self) as usize;
+		//  If there are `n` bits live between the head cursor (which marks the
 		//  address of the first live bit) and the back edge of the element,
-		//  then when len is <= n, the span covers one element.
-		//  When len == n, the tail will be T::SIZE, which is valid for a tail.
-		//  TODO(myrrlyn): Separate BitIdx into Head and Tail types, which have
-		//  their proper range enforcements.
+		//  then when `len <= n`, the span covers one element. When `len == n`,
+		//  the tail will be `T::BITS`, which is valid for a tail.
 		if len <= bits_in_head {
 			(1, (*self + len as u8).into())
 		}
-		//  If there are more bits in the span than n, then subtract n from len
-		//  and use the difference to count elements and bits.
+		//  If there are more bits in the span than `n`, then subtract `n` from
+		//  `len` and use the difference to count elements and bits.
 		else {
 			//  1 ..
 			let bits_after_head = len - bits_in_head;
 			//  Count the number of wholly filled elements
-			let whole_elts = bits_after_head >> T::BITS;
+			let whole_elts = bits_after_head >> T::INDX;
 			//  Count the number of bits in the *next* element. If this is zero,
-			//  become T::SIZE; if it is nonzero, add one more to elts. elts
-			//  must have one added to it by default to account for the head
-			//  element.
+			//  become `T::BITS`; if it is nonzero, add one more to `elts`.
+			//  `elts` must have one added to it by default to account for the
+			//  head element.
 			let tail_bits = bits_after_head as u8 & T::MASK;
 			if tail_bits == 0 {
-				(whole_elts + 1, T::SIZE.into())
+				(whole_elts + 1, T::BITS.into())
 			}
 			else {
 				(whole_elts + 2, tail_bits.into())
@@ -603,10 +1019,16 @@ impl Display for BitIdx {
 
 impl Deref for BitIdx {
 	type Target = u8;
-	fn deref(&self) -> &Self::Target { &self.0 } }
+
+	fn deref(&self) -> &Self::Target {
+		&self.0
+	}
+}
 
 impl DerefMut for BitIdx {
-	fn deref_mut(&mut self) -> &mut Self::Target { &mut self.0 }
+	fn deref_mut(&mut self) -> &mut Self::Target {
+		&mut self.0
+	}
 }
 
 /** Newtype indicating a concrete index into an element.
@@ -617,9 +1039,10 @@ an element rather than a semantic bit.
 `Cursor` implementors translate `BitIdx` values, which are semantic places, into
 `BitPos` values, which are concrete electrical positions.
 
-[`Cursor`]: ../trait.Cursor.html
+[`Cursor`]: ../cursor/trait.Cursor.html
 **/
 #[derive(Clone, Copy, Debug, Default, Eq, Hash, Ord, PartialEq, PartialOrd)]
+#[doc(hidden)]
 pub struct BitPos(u8);
 
 impl BitPos {
@@ -636,8 +1059,9 @@ impl BitPos {
 	/// # Type Parameters
 	///
 	/// - `T: Bits`: The storage type used to determine position validity.
-	pub fn is_valid<T: Bits>(self) -> bool {
-		*self < T::SIZE
+	pub fn is_valid<T>(self) -> bool
+	where T: Bits {
+		self.load() < T::BITS
 	}
 }
 
@@ -663,19 +1087,46 @@ impl Display for BitPos {
 
 impl Deref for BitPos {
 	type Target = u8;
-	fn deref(&self) -> &Self::Target { &self.0 }
+
+	fn deref(&self) -> &Self::Target {
+		&self.0
+	}
 }
 
 impl DerefMut for BitPos {
-	fn deref_mut(&mut self) -> &mut Self::Target { &mut self.0 }
+	fn deref_mut(&mut self) -> &mut Self::Target {
+		&mut self.0
+	}
 }
 
-impl Bits for u8  { const BITS: u8 = 3; const TYPENAME: &'static str = "u8";  }
-impl Bits for u16 { const BITS: u8 = 4; const TYPENAME: &'static str = "u16"; }
-impl Bits for u32 { const BITS: u8 = 5; const TYPENAME: &'static str = "u32"; }
+impl Bits for u8 {
+	const TYPENAME: &'static str = "u8";
+
+	#[cfg(feature = "atomic")]
+	type Atom = atomic::AtomicU8;
+}
+
+impl Bits for u16 {
+	const TYPENAME: &'static str = "u16";
+
+	#[cfg(feature = "atomic")]
+	type Atom = atomic::AtomicU16;
+}
+
+impl Bits for u32 {
+	const TYPENAME: &'static str = "u32";
+
+	#[cfg(feature = "atomic")]
+	type Atom = atomic::AtomicU32;
+}
 
 #[cfg(target_pointer_width = "64")]
-impl Bits for u64 { const BITS: u8 = 6; const TYPENAME: &'static str = "u64"; }
+impl Bits for u64 {
+	const TYPENAME: &'static str = "u64";
+
+	#[cfg(feature = "atomic")]
+	type Atom = atomic::AtomicU64;
+}
 
 /// Marker trait to seal `Bits` against downstream implementation.
 ///
@@ -703,11 +1154,11 @@ mod tests {
 		//  than the start bit.
 		for n in 1 .. 8 {
 			let (elt, bit) = BitIdx::from(n).offset::<u8>(isize::max_value());
-			assert_eq!(elt, (isize::max_value() >> u8::BITS) + 1);
+			assert_eq!(elt, (isize::max_value() >> u8::INDX) + 1);
 			assert_eq!(*bit, n - 1);
 		}
 		let (elt, bit) = BitIdx::from(0).offset::<u8>(isize::max_value());
-		assert_eq!(elt, isize::max_value() >> u8::BITS);
+		assert_eq!(elt, isize::max_value() >> u8::INDX);
 		assert_eq!(*bit, 7);
 	}
 
@@ -717,7 +1168,7 @@ mod tests {
 		//  the start bit
 		for n in 0 .. 8 {
 			let (elt, bit) = BitIdx::from(n).offset::<u8>(isize::min_value());
-			assert_eq!(elt, isize::min_value() >> u8::BITS);
+			assert_eq!(elt, isize::min_value() >> u8::INDX);
 			assert_eq!(*bit, n);
 		}
 	}
