@@ -4,26 +4,29 @@ This module holds the type for an owned but ungrowable bit sequence. `BitVec` is
 the more appropriate and useful type for most collections.
 !*/
 
-#![cfg(feature = "alloc")]
+#![cfg(any(feature = "alloc", feature = "std"))]
 
 use crate::{
-	BigEndian,
-	BitPtr,
-	BitSlice,
-	BitVec,
-	Bits,
-	Cursor,
+	bits::Bits,
+	cursor::{
+		BigEndian,
+		Cursor,
+	},
+	pointer::BitPtr,
+	slice::BitSlice,
+	vec::BitVec,
 };
+
 #[cfg(all(feature = "alloc", not(feature = "std")))]
 use alloc::{
 	borrow::{
 		Borrow,
 		BorrowMut,
-		ToOwned,
 	},
 	boxed::Box,
 	vec::Vec,
 };
+
 use core::{
 	clone::Clone,
 	cmp::{
@@ -90,12 +93,12 @@ use core::{
 		ShrAssign,
 	},
 };
+
 #[cfg(feature = "std")]
 use std::{
 	borrow::{
 		Borrow,
 		BorrowMut,
-		ToOwned,
 	},
 	boxed::Box,
 	vec::Vec,
@@ -137,40 +140,168 @@ where C: Cursor, T: Bits {
 
 impl<C, T> BitBox<C, T>
 where C: Cursor, T: Bits {
-	/// Constructs an empty slice at a given location.
-	///
-	/// # Parameters
-	///
-	/// - `data`: The address of the empty `BitBox` to construct.
+	/// Constructs an empty boxed bitslice.
 	///
 	/// # Returns
 	///
-	/// An empty `BitBox` at the given location.
-	pub fn uninhabited(data: *const T) -> Self {
+	/// An empty `BitBox` at an arbitrary location.
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use bitvec::prelude::*;
+	///
+	/// let bb: BitBox = BitBox::empty();
+	/// assert!(bb.is_empty());
+	/// ```
+	pub fn empty() -> Self {
 		Self {
 			_cursor: PhantomData,
-			pointer: BitPtr::uninhabited(data),
+			pointer: BitPtr::empty(),
 		}
 	}
 
-	/// Copies a `BitSlice` into an owned `BitBox`.
+	/// Produces a `BitBox` from a single element.
 	///
 	/// # Parameters
 	///
-	/// - `src`: The `&BitSlice` to make owned.
+	/// - `elt`: The source element from which to make the `BitBox`.
 	///
 	/// # Returns
 	///
-	/// An owned clone of the given bit slice.
-	pub fn new(src: &BitSlice<C, T>) -> Self {
-		let store: Box<[T]> = src.as_ref().to_owned().into_boxed_slice();
-		let data = store.as_ptr();
-		let (_, elts, head, tail) = src.bitptr().raw_parts();
+	/// A `BitBox` containing the provided element.
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use bitvec::prelude::*;
+	///
+	/// let bb: BitBox<BigEndian, u16> = BitBox::from_element(!0);
+	/// assert!(bb.all());
+	/// ```
+	pub fn from_element(elt: T) -> Self {
+		BitSlice::<C, T>::from_element(&elt).into()
+	}
+
+	/// Builds a `BitBox` from a borrowed slice of elements.
+	///
+	/// # Parameters
+	///
+	/// - `slice`: The source slice from which to make the `BitBox`.
+	///
+	/// # Returns
+	///
+	/// A `BitBox` containing the (cloned) provided slice.
+	///
+	/// # Panics
+	///
+	/// This function may panic if the provided slice is longer than the
+	/// `BitBox` can support.
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use bitvec::prelude::*;
+	///
+	/// let src: &[u8] = &[5, 10];
+	/// let bv: BitBox = src.into();
+	/// assert!(bv[5]);
+	/// assert!(bv[7]);
+	/// assert!(bv[12]);
+	/// assert!(bv[14]);
+	/// ```
+	pub fn from_slice(slice: &[T]) -> Self {
+		assert!(slice.len() <= BitPtr::<T>::MAX_ELTS, "Box overflow");
+		BitVec::from_slice(slice).into_boxed_bitslice()
+	}
+
+	/// Clones a `&BitSlice` into a `BitBox`.
+	///
+	/// # Parameters
+	///
+	/// - `slice`: The bit slice to clone into a bit box.
+	///
+	/// # Returns
+	///
+	/// A `BitBox` containing the same bits as the source slice.
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use bitvec::prelude::*;
+	///
+	/// let src: &[u8] = &[0, !0];
+	/// let bs: &BitSlice = src.into();
+	/// let bb = BitBox::from_bitslice(bs);
+	/// assert_eq!(bb.len(), 16);
+	/// assert!(bb.some());
+	/// ```
+	pub fn from_bitslice(slice: &BitSlice<C, T>) -> Self {
+		BitVec::from_bitslice(slice).into_boxed_bitslice()
+	}
+
+	/// Produces a `BitBox` from an owned slice of elements.
+	///
+	/// # Parameters
+	///
+	/// - `slice`: The source boxed slice from which to make the `BitBox`.
+	///
+	/// # Returns
+	///
+	/// A `BitBox` governing the same slice that was passed in. This function
+	/// does not reallocate.
+	///
+	/// # Panics
+	///
+	/// This function may panic if the provided slice is longer than the
+	/// `BitBox` can support.
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use bitvec::prelude::*;
+	///
+	/// let slice: Box<[u16]> = vec![0, !0].into_boxed_slice();
+	/// let bb = BitBox::<LittleEndian, _>::from_boxed_slice(slice);
+	/// assert!(bb.some());
+	/// assert_eq!(bb.len(), 32);
+	/// ```
+	pub fn from_boxed_slice(slice: Box<[T]>) -> Self {
+		assert!(slice.len() <= BitPtr::<T>::MAX_ELTS, "Box overflow");
 		let out = Self {
 			_cursor: PhantomData,
-			pointer: BitPtr::new(data, elts, head, tail),
+			pointer: BitPtr::new(slice.as_ptr(), slice.len(), 0, T::BITS)
 		};
-		mem::forget(store);
+		mem::forget(slice);
+		out
+	}
+
+	/// Removes the `BitBox` wrapper from a `Box<[T]>`.
+	///
+	/// # Parameters
+	///
+	/// - `self`
+	///
+	/// # Returns
+	///
+	/// The `Box<[T]>` underneath `self`.
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use bitvec::prelude::*;
+	///
+	/// let slice: Box<[u16]> = vec![0, !0].into_boxed_slice();
+	/// let bb = BitBox::<LittleEndian, _>::from_boxed_slice(slice);
+	/// assert_eq!(bb.len(), 32);
+	/// let slice = bb.into_boxed_slice();
+	/// assert_eq!(slice.len(), 2);
+	/// ```
+	pub fn into_boxed_slice(self) -> Box<[T]> {
+		let (ptr, len, _, _) = self.bitptr().raw_parts();
+		let out = unsafe { Vec::from_raw_parts(ptr as *mut T, len, len) }
+			.into_boxed_slice();
+		mem::forget(self);
 		out
 	}
 
@@ -215,17 +346,10 @@ where C: Cursor, T: Bits {
 	/// The proper way to do so is to convert the `BitPtr` back into a `BitBox`
 	/// with the [`BitBox::from_raw`] function.
 	///
-	/// Note: this is an associated function, which means that you must call it
-	/// as `BitBox::into_raw(b)` instead of `b.into_raw()`. This is to match the
-	/// API of [`Box`]; there is no method conflict with [`BitSlice`].
-	///
 	/// [`BitBox::from_raw`]: #method.from_raw
-	/// [`BitSlice`]: ../struct.BitSlice.html
-	#[cfg_attr(not(feature = "std"), doc = "[`Box`]: https://doc.rust-lang.org/stable/alloc/boxed/struct.Box.html")]
-	#[cfg_attr(feature = "std", doc = "[`Box`]: https://doc.rust-lang.org/stable/std/boxed/struct.Box.html")]
-	pub unsafe fn into_raw(b: BitBox<C, T>) -> BitPtr<T> {
-		let out = b.bitptr();
-		mem::forget(b);
+	pub unsafe fn into_raw(self) -> BitPtr<T> {
+		let out = self.bitptr();
+		mem::forget(self);
 		out
 	}
 
@@ -240,28 +364,39 @@ where C: Cursor, T: Bits {
 	/// `BitBox`. This `BitBox` can then be dropped which will properly
 	/// deallocate the memory.
 	///
-	/// Note: this is an associated function, which means that you must call it
-	/// as `BitBox::leak(b)` instead of `b.leak()`. This is to match the API of
-	/// [`Box`]; there is no method conflict with [`BitSlice`].
-	///
 	/// # Parameters
 	///
 	/// - `b`: The `BitBox` to deconstruct.
 	///
 	/// # Returns
 	///
-	/// The raw pointer from inside the `BitBox`.
+	/// The slice formerly governed by the `BitBox`, which will never
+	/// deallocate.
 	///
 	/// [`BitBox::from_raw`]: #method.from_raw
-	/// [`BitSlice`]: ../struct.BitSlice.html
-	#[cfg_attr(not(feature = "std"), doc = "[`Box`]: https://doc.rust-lang.org/stable/alloc/boxed/struct.Box.html")]
-	#[cfg_attr(feature = "std", doc = "[`Box`]: https://doc.rust-lang.org/stable/std/boxed/struct.Box.html")]
-	pub fn leak<'a>(b: BitBox<C, T>) -> &'a mut BitSlice<C, T> {
-		let out = b.bitptr();
-		mem::forget(b);
+	pub fn leak<'a>(self) -> &'a mut BitSlice<C, T> {
+		let out = self.bitptr();
+		mem::forget(self);
 		out.into()
 	}
 
+	/// Changes the cursor on a box handle, without changing the data it
+	/// governs.
+	///
+	/// # Parameters
+	///
+	/// - `self`
+	///
+	/// # Returns
+	///
+	/// An equivalent handle to the same data, with a new cursor parameter.
+	pub fn change_cursor<D>(self) -> BitBox<D, T>
+	where D: Cursor {
+		let bp = self.bitptr();
+		mem::forget(self);
+		unsafe { BitBox::from_raw(bp) }
+	}
+
 	/// Accesses the `BitSlice<C, T>` to which the `BitBox` refers.
 	///
 	/// # Parameters
@@ -347,7 +482,7 @@ where C: Cursor, T: Bits {
 impl<C, T> Clone for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn clone(&self) -> Self {
-		let (_, e, h, t) = self.bitptr().raw_parts();
+		let (e, h, t) = self.bitptr().region_data();
 		let new_box = self.do_with_box(Clone::clone);
 		let ptr = new_box.as_ptr();
 		mem::forget(new_box);
@@ -364,49 +499,49 @@ where C: Cursor, T: Bits {}
 impl<C, T> Ord for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn cmp(&self, rhs: &Self) -> Ordering {
-		(&**self).cmp(&**rhs)
+		self.as_bitslice().cmp(rhs.as_bitslice())
 	}
 }
 
 impl<A, B, C, D> PartialEq<BitBox<C, D>> for BitBox<A, B>
 where A: Cursor, B: Bits, C: Cursor, D: Bits {
 	fn eq(&self, rhs: &BitBox<C, D>) -> bool {
-		(&**self).eq(&**rhs)
+		self.as_bitslice().eq(rhs.as_bitslice())
 	}
 }
 
 impl<A, B, C, D> PartialEq<BitSlice<C, D>> for BitBox<A, B>
 where A: Cursor, B: Bits, C: Cursor, D: Bits {
 	fn eq(&self, rhs: &BitSlice<C, D>) -> bool {
-		(&**self).eq(rhs)
+		self.as_bitslice().eq(rhs)
 	}
 }
 
 impl<A, B, C, D> PartialEq<BitBox<C, D>> for BitSlice<A, B>
 where A: Cursor, B: Bits, C: Cursor, D: Bits {
 	fn eq(&self, rhs: &BitBox<C, D>) -> bool {
-		self.eq(&**rhs)
+		self.eq(rhs.as_bitslice())
 	}
 }
 
 impl<A, B, C, D> PartialOrd<BitBox<C, D>> for BitBox<A, B>
 where A: Cursor, B: Bits, C: Cursor, D: Bits {
 	fn partial_cmp(&self, rhs: &BitBox<C, D>) -> Option<Ordering> {
-		(&**self).partial_cmp(&**rhs)
+		self.as_bitslice().partial_cmp(rhs.as_bitslice())
 	}
 }
 
 impl<A, B, C, D> PartialOrd<BitSlice<C, D>> for BitBox<A, B>
 where A: Cursor, B: Bits, C: Cursor, D: Bits {
 	fn partial_cmp(&self, rhs: &BitSlice<C, D>) -> Option<Ordering> {
-		(&**self).partial_cmp(rhs)
+		self.as_bitslice().partial_cmp(rhs)
 	}
 }
 
 impl<A, B, C, D> PartialOrd<BitBox<C, D>> for BitSlice<A, B>
 where A: Cursor, B: Bits, C: Cursor, D: Bits {
 	fn partial_cmp(&self, rhs: &BitBox<C, D>) -> Option<Ordering> {
-		self.partial_cmp(&**rhs)
+		self.partial_cmp(rhs.as_bitslice())
 	}
 }
 
@@ -420,7 +555,7 @@ where C: Cursor, T: Bits {
 impl<C, T> AsMut<[T]> for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn as_mut(&mut self) -> &mut [T] {
-		(&mut **self).as_mut()
+		self.as_mut_bitslice().as_mut()
 	}
 }
 
@@ -434,113 +569,42 @@ where C: Cursor, T: Bits {
 impl<C, T> AsRef<[T]> for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn as_ref(&self) -> &[T] {
-		(&**self).as_ref()
+		self.as_bitslice().as_ref()
 	}
 }
 
 impl<C, T> From<&BitSlice<C, T>> for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn from(src: &BitSlice<C, T>) -> Self {
-		let (_, elts, head, tail) = src.bitptr().raw_parts();
-		let b: Box<[T]> = src.as_ref().to_owned().into_boxed_slice();
-		let out = Self {
-			_cursor: PhantomData,
-			pointer: BitPtr::new(b.as_ptr(), elts, head, tail),
-		};
-		mem::forget(b);
-		out
+		Self::from_bitslice(src)
 	}
 }
 
-/// Builds a `BitBox` out of a borrowed slice of elements.
-///
-/// This copies the memory as-is from the source buffer into the new `BitBox`.
-/// The source buffer will be unchanged by this operation, so you don't need to
-/// worry about using the correct cursor type for the read.
-///
-/// This operation does a copy from the source buffer into a new allocation, as
-/// it can only borrow the source and not take ownership.
 impl<C, T> From<&[T]> for BitBox<C, T>
 where C: Cursor, T: Bits {
-	/// Builds a `BitBox<C: Cursor, T: Bits>` from a borrowed `&[T]`.
-	///
-	/// # Parameters
-	///
-	/// - `src`: The elements to use as the values for the new vector.
-	///
-	/// # Examples
-	///
-	/// ```rust
-	/// use bitvec::*;
-	///
-	/// let src: &[u8] = &[5, 10];
-	/// let bv: BitBox = src.into();
-	/// assert!(bv[5]);
-	/// assert!(bv[7]);
-	/// assert!(bv[12]);
-	/// assert!(bv[14]);
-	/// ```
 	fn from(src: &[T]) -> Self {
-		assert!(src.len() < BitPtr::<T>::MAX_ELTS, "Box overflow");
-		<&BitSlice<C, T>>::from(src).into()
+		Self::from_slice(src)
 	}
 }
 
 impl<C, T> From<BitVec<C, T>> for BitBox<C, T>
 where C: Cursor, T: Bits {
-	fn from(mut src: BitVec<C, T>) -> Self {
-		src.shrink_to_fit();
-		let pointer = src.bitptr();
-		mem::forget(src);
-		unsafe { Self::from_raw(pointer) }
+	fn from(src: BitVec<C, T>) -> Self {
+		src.into_boxed_bitslice()
 	}
 }
 
-/// Builds a `BitBox` out of an owned slice of elements.
-///
-/// This moves the memory as-is from the source buffer into the new `BitBox`.
-/// The source buffer will be unchanged by this operation, so you don't need to
-/// worry about using the correct cursor type.
 impl<C, T> From<Box<[T]>> for BitBox<C, T>
 where C: Cursor, T: Bits {
-	/// Consumes a `Box<[T: Bits]>` and creates a `BitBox<C: Cursor, T>` from
-	/// it.
-	///
-	/// # Parameters
-	///
-	/// - `src`: The source box whose memory will be used.
-	///
-	/// # Returns
-	///
-	/// A new `BitBox` using the `src` `Box`’s memory.
-	///
-	/// # Examples
-	///
-	/// ```rust
-	/// use bitvec::*;
-	///
-	/// let src: Box<[u8]> = Box::new([3, 6, 9, 12, 15]);
-	/// let bv: BitBox = src.into();
-	/// ```
 	fn from(src: Box<[T]>) -> Self {
-		assert!(src.len() < BitPtr::<T>::MAX_ELTS, "Box overflow");
-		let out = Self {
-			_cursor: PhantomData,
-			pointer: BitPtr::new(src.as_ptr(), src.len(), 0, T::SIZE)
-		};
-		mem::forget(src);
-		out
+		Self::from_boxed_slice(src)
 	}
 }
 
 impl<C, T> Into<Box<[T]>> for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn into(self) -> Box<[T]> {
-		let (ptr, len, _, _) = self.bitptr().raw_parts();
-		let out = unsafe { Vec::from_raw_parts(ptr as *mut T, len, len) }
-			.into_boxed_slice();
-		mem::forget(self);
-		out
+		self.into_boxed_slice()
 	}
 }
 
@@ -576,7 +640,7 @@ where C: Cursor, T: Bits {
 impl<C, T> Hash for BitBox<C, T>
 where C: Cursor, T: Bits {
 	fn hash<H: Hasher>(&self, hasher: &mut H) {
-		(&**self).hash(hasher)
+		self.as_bitslice().hash(hasher)
 	}
 }
 
@@ -599,7 +663,7 @@ where C: Cursor, T: 'a + Bits {
 	type IntoIter = <&'a BitSlice<C, T> as IntoIterator>::IntoIter;
 
 	fn into_iter(self) -> Self::IntoIter {
-		(&**self).into_iter()
+		self.as_bitslice().into_iter()
 	}
 }
 
