@@ -20,10 +20,10 @@ system.
 Like `vec!`, `bitvec!` supports bit lists `[0, 1, …]` and repetition markers
 `[1; n]`.
 
-# All Syntaxes
+# Examples
 
 ```rust
-use bitvec::*;
+use bitvec::prelude::*;
 
 bitvec![BigEndian, u8; 0, 1];
 bitvec![LittleEndian, u8; 0, 1,];
@@ -36,7 +36,7 @@ bitvec![LittleEndian; 0; 5];
 bitvec![1; 5];
 ```
 **/
-#[cfg(feature = "alloc")]
+#[cfg(any(feature = "alloc", feature = "std"))]
 #[macro_export]
 macro_rules! bitvec {
 	//  bitvec![ endian , type ; 0 , 1 , … ]
@@ -59,11 +59,11 @@ macro_rules! bitvec {
 
 	//  bitvec![ 0 , 1 , … ]
 	( $( $element:expr ),* ) => {
-		bitvec![ __bv_impl__ $crate::BigEndian , u8 ; $( $element ),* ]
+		bitvec![ __bv_impl__ $crate::prelude::BigEndian , u8 ; $( $element ),* ]
 	};
 	//  bitvec![ 0 , 1 , … , ]
 	( $( $element:expr , )* ) => {
-		bitvec![ __bv_impl__ $crate::BigEndian , u8 ; $( $element ),* ]
+		bitvec![ __bv_impl__ $crate::prelude::BigEndian , u8 ; $( $element ),* ]
 	};
 
 	//  bitvec![ endian , type ; bit ; rep ]
@@ -76,7 +76,7 @@ macro_rules! bitvec {
 	};
 	//  bitvec![ bit ; rep ]
 	( $element:expr ; $rep:expr ) => {
-		bitvec![ __bv_impl__ $crate::BigEndian , u8 ; $element ; $rep ]
+		bitvec![ __bv_impl__ $crate::prelude::BigEndian , u8 ; $element ; $rep ]
 	};
 
 	//  Build an array of `bool` (one bit per byte) and then build a `BitVec`
@@ -89,95 +89,159 @@ macro_rules! bitvec {
 
 	( __bv_impl__ $endian:path , $bits:ty ; $( $element:expr ),* ) => {{
 		let init: &[bool] = &[ $( $element != 0 ),* ];
-		$crate :: BitVec :: < $endian , $bits > :: from ( init )
+		$crate :: vec :: BitVec :: < $endian , $bits > :: from ( init )
 	}};
 
 	( __bv_impl__ $endian:path , $bits:ty ; $element:expr ; $rep:expr ) => {{
 		core :: iter :: repeat ( $element != 0 )
 			.take ( $rep )
-			.collect :: < $crate :: BitVec < $endian , $bits > > ( )
+			.collect :: < $crate :: vec :: BitVec < $endian , $bits > > ( )
 	}};
 }
 
+/** Construct a `BitBox` out of a literal array in source code, like `bitvec!`.
+
+This has exactly the same syntax as [`bitvec!`], and in fact is a thin wrapper
+around `bitvec!` that calls `.into_boxed_slice()` on the produced `BitVec` to
+freeze it.
+
+[`bitvec!`]: #macro.bitvec
+**/
+#[cfg(any(feature = "alloc", feature = "std"))]
+#[macro_export]
+macro_rules! bitbox {
+	//  bitbox![ endian , type ; 0 , 1 , … ]
+	( $endian:path , $bits:ty ; $( $element:expr ),* ) => {
+		bitvec![ $endian , $bits ; $( $element ),* ].into_boxed_bitslice()
+	};
+	//  bitbox![ endian , type ; 0 , 1 , … , ]
+	( $endian:path , $bits:ty ; $( $element:expr , )* ) => {
+		bitvec![ $endian , $bits ; $( $element ),* ].into_boxed_bitslice()
+	};
+
+	//  bitbox![ endian ; 0 , 1 , … ]
+	( $endian:path ; $( $element:expr ),* ) => {
+		bitvec![ $endian , u8 ; $( $element ),* ].into_boxed_bitslice()
+	};
+	//  bitbox![ endian ; 0 , 1 , … , ]
+	( $endian:path ; $( $element:expr , )* ) => {
+		bitvec![ $endian , u8 ; $( $element ),* ].into_boxed_bitslice()
+	};
+
+	//  bitbox![ 0 , 1 , … ]
+	( $( $element:expr ),* ) => {
+		bitvec![ $crate::prelude::BigEndian , u8 ; $( $element ),* ].into_boxed_bitslice()
+	};
+	//  bitbox![ 0 , 1 , … , ]
+	( $( $element:expr , )* ) => {
+		bitvec![ $crate::prelude::BigEndian , u8 ; $( $element ),* ].into_boxed_bitslice()
+	};
+
+	//  bitbox![ endian , type ; bit ; rep ]
+	( $endian:path , $bits:ty ; $element:expr ; $rep:expr ) => {
+		bitvec![ $endian , $bits ; $element; $rep ].into_boxed_bitslice()
+	};
+	//  bitbox![ endian ; bit ; rep ]
+	( $endian:path ; $element:expr ; $rep:expr ) => {
+		bitvec![ $endian , u8 ; $element ; $rep ].into_boxed_bitslice()
+	};
+	//  bitbox![ bit ; rep ]
+	( $element:expr ; $rep:expr ) => {
+		bitvec![ $crate::prelude::BigEndian , u8 ; $element ; $rep ].into_boxed_bitslice()
+	};
+}
+
 #[doc(hidden)]
 macro_rules! __bitslice_shift {
 	( $( $t:ty ),+ ) => { $(
 		#[doc(hidden)]
-		impl<C: $crate :: Cursor, T: $crate :: Bits> core::ops::ShlAssign< $t >
-		for $crate :: BitSlice<C, T>
-		{
-			fn shl_assign(&mut self, shamt: $t ) {
-				core::ops::ShlAssign::<usize>::shl_assign(self, shamt as usize);
+		impl < C , T > core :: ops :: ShlAssign < $t >
+		for $crate :: prelude :: BitSlice < C , T >
+		where C : $crate :: cursor :: Cursor , T : $crate :: bits :: Bits {
+			fn shl_assign ( & mut self , shamt : $t ) {
+				core :: ops :: ShlAssign :: < usize > :: shl_assign (
+					self ,
+					shamt as usize ,
+				)
 			}
 		}
 
 		#[doc(hidden)]
-		impl<C: $crate :: Cursor, T: $crate :: Bits> core::ops::ShrAssign< $t >
-		for $crate :: BitSlice<C, T>
-		{
-			fn shr_assign(&mut self, shamt: $t ) {
-				core::ops::ShrAssign::<usize>::shr_assign(self, shamt as usize);
+		impl < C , T > core :: ops :: ShrAssign < $t >
+		for $crate :: prelude :: BitSlice < C , T >
+		where C : $crate :: cursor :: Cursor , T : $crate :: bits :: Bits {
+			fn shr_assign ( & mut self , shamt : $t ) {
+				core :: ops :: ShrAssign :: < usize > :: shr_assign (
+					self,
+					shamt as usize ,
+				)
 			}
 		}
 	)+ };
 }
 
-#[cfg(feature = "alloc")]
+#[cfg(any(feature = "alloc", feature = "std"))]
 #[doc(hidden)]
 macro_rules! __bitvec_shift {
 	( $( $t:ty ),+ ) => { $(
 		#[doc(hidden)]
-		impl<C: $crate :: Cursor, T: $crate :: Bits> core::ops::Shl< $t >
-		for $crate ::BitVec<C, T>
-		{
-			type Output = <Self as core::ops::Shl<usize>>::Output;
+		impl < C , T > core :: ops :: Shl < $t >
+		for $crate :: vec :: BitVec < C , T >
+		where C : $crate :: cursor :: Cursor , T : $crate :: bits :: Bits {
+			type Output = < Self as core :: ops :: Shl < usize > > :: Output ;
 
-			fn shl(self, shamt: $t ) -> Self::Output {
-				core::ops::Shl::<usize>::shl(self, shamt as usize)
+			fn shl ( self , shamt : $t ) -> Self :: Output {
+				core :: ops :: Shl :: < usize > :: shl ( self , shamt as usize )
 			}
 		}
 
 		#[doc(hidden)]
-		impl<C: $crate :: Cursor, T: $crate :: Bits> core::ops::ShlAssign< $t >
-		for $crate ::BitVec<C, T>
-		{
-			fn shl_assign(&mut self, shamt: $t ) {
-				core::ops::ShlAssign::<usize>::shl_assign(self, shamt as usize)
+		impl < C , T > core :: ops :: ShlAssign < $t >
+		for $crate :: vec :: BitVec < C , T >
+		where C : $crate :: cursor :: Cursor , T : $crate :: bits :: Bits {
+			fn shl_assign ( & mut self , shamt : $t ) {
+				core :: ops :: ShlAssign :: < usize > :: shl_assign (
+					self ,
+					shamt as usize,
+				)
 			}
 		}
 
 		#[doc(hidden)]
-		impl<C: $crate ::Cursor, T: $crate ::Bits> core::ops::Shr< $t >
-		for $crate ::BitVec<C, T>
-		{
-			type Output = <Self as core::ops::Shr<usize>>::Output;
+		impl < C , T > core :: ops :: Shr < $t >
+		for $crate :: vec :: BitVec < C , T >
+		where C : $crate :: cursor :: Cursor , T : $crate :: bits :: Bits {
+			type Output = < Self as core :: ops :: Shr < usize > > :: Output ;
 
-			fn shr(self, shamt: $t ) -> Self::Output {
-				core::ops::Shr::<usize>::shr(self, shamt as usize)
+			fn shr ( self , shamt : $t ) -> Self :: Output {
+				core :: ops :: Shr :: < usize > :: shr ( self , shamt as usize )
 			}
 		}
 
 		#[doc(hidden)]
-		impl<C: $crate ::Cursor, T: $crate ::Bits> core::ops::ShrAssign< $t >
-		for $crate ::BitVec<C, T>
-		{
-			fn shr_assign(&mut self, shamt: $t ) {
-				core::ops::ShrAssign::<usize>::shr_assign(self, shamt as usize)
+		impl < C , T> core :: ops :: ShrAssign < $t >
+		for $crate :: vec :: BitVec < C , T >
+		where C : $crate :: cursor :: Cursor , T : $crate :: bits :: Bits {
+			fn shr_assign ( & mut self , shamt : $t ) {
+				core :: ops :: ShrAssign :: < usize > :: shr_assign (
+					self ,
+					shamt as usize ,
+				)
 			}
 		}
 	)+ };
 }
 
-#[cfg(all(test, feature = "alloc"))]
+#[cfg(all(test, any(feature = "alloc", feature = "std")))]
 mod tests {
 	#[allow(unused_imports)]
-	use crate::{
+	use crate::cursor::{
 		BigEndian,
 		LittleEndian,
 	};
 
 	#[test]
-	fn compile_macros() {
+	fn compile_bitvec_macros() {
 		bitvec![0, 1];
 		bitvec![BigEndian; 0, 1];
 		bitvec![LittleEndian; 0, 1];
@@ -202,4 +266,31 @@ mod tests {
 		bitvec![BigEndian, u64; 0; 70];
 		bitvec![LittleEndian, u64; 1; 70];
 	}
+
+	#[test]
+	fn compile_bitbox_macros() {
+		bitbox![0, 1];
+		bitbox![BigEndian; 0, 1];
+		bitbox![LittleEndian; 0, 1];
+		bitbox![BigEndian, u8; 0, 1];
+		bitbox![LittleEndian, u8; 0, 1];
+		bitbox![BigEndian, u16; 0, 1];
+		bitbox![LittleEndian, u16; 0, 1];
+		bitbox![BigEndian, u32; 0, 1];
+		bitbox![LittleEndian, u32; 0, 1];
+		bitbox![BigEndian, u64; 0, 1];
+		bitbox![LittleEndian, u64; 0, 1];
+
+		bitbox![1; 70];
+		bitbox![BigEndian; 0; 70];
+		bitbox![LittleEndian; 1; 70];
+		bitbox![BigEndian, u8; 0; 70];
+		bitbox![LittleEndian, u8; 1; 70];
+		bitbox![BigEndian, u16; 0; 70];
+		bitbox![LittleEndian, u16; 1; 70];
+		bitbox![BigEndian, u32; 0; 70];
+		bitbox![LittleEndian, u32; 1; 70];
+		bitbox![BigEndian, u64; 0; 70];
+		bitbox![LittleEndian, u64; 1; 70];
+	}
 }
