@@ -7,11 +7,15 @@ work.
 !*/
 
 use crate::{
-	BitIdx,
-	BitSlice,
-	Bits,
-	Cursor,
+	bits::{
+		BitIdx,
+		Bits,
+	},
+	cursor::Cursor,
+	domain::*,
+	slice::BitSlice,
 };
+
 use core::{
 	convert::{
 		AsMut,
@@ -25,17 +29,16 @@ use core::{
 		Formatter,
 	},
 	marker::PhantomData,
-	mem,
+	mem::size_of,
 	ptr::NonNull,
 	slice,
 };
 
-
 /// Width in bits of a pointer on the target machine.
-const PTR_BITS: usize = mem::size_of::<*const u8>() * 8;
+const PTR_BITS: usize = size_of::<*const u8>() * 8;
 
 /// Width in bits of a processor word on the target machine.
-const USZ_BITS: usize = mem::size_of::<usize>() * 8;
+const USZ_BITS: usize = size_of::<usize>() * 8;
 
 /** In-memory representation of `&BitSlice` handles.
 
@@ -47,8 +50,8 @@ counter fundamentals into sub-field components. Rust does not have bitfield
 syntax, so the below description of the element layout is in C++.
 
 ```cpp
-template<typename T>
-struct BitPtr<T> {
+template <typename T>
+struct BitPtr {
   size_t ptr_head : __builtin_ctzll(alignof(T));
   size_t ptr_data : sizeof(T*) * 8
                   - __builtin_ctzll(alignof(T));
@@ -123,18 +126,9 @@ The following value sets are edge cases of valid `BitPtr` structures.
 
 ## Empty Slice
 
-The empty slice is canonically represented by a wholly zeroed slot:
-
-- `data`: `core::ptr::null::<T>()`
-- `elts`: `0usize`
-- `head`: `0u8`
-- `tail`: `0u8`
-- `ptr`: `core::ptr::null::<u8>()`
-- `len`: `0usize`
-
-All `BitPtr` values whose `data` pointer is `null` represents the empty slice,
-regardless of other field contents, but the normalized form zeros all other
-fields also.
+The empty slice has all its fields zeroed except for the data pointer, which is
+any non-null pointer. The null pointer value is used to indicate
+`Option::<BitPtr<T>::None`.
 
 ## Allocated, Uninhabited, Slice
 
@@ -142,7 +136,7 @@ An allocated, owned, region of memory that is uninhabited. This is functionally
 the empty slice, but it must retain its pointer information. All other fields in
 the slot are zeroed.
 
-- `data`: (any valid `*const T`)
+- `data`: (any valid `*const T` other than `NonNull::<T>::dangling()`)
 - `elts`: `0usize`
 - `head`: `0u8`
 - `tail`: `0u8`
@@ -177,19 +171,21 @@ regime.
 **/
 #[repr(C)]
 #[derive(Clone, Copy, Eq, Hash, PartialEq, PartialOrd, Ord)]
-pub struct BitPtr<T>
+pub struct BitPtr<T = u8>
 where T: Bits {
 	_ty: PhantomData<T>,
-	/// Pointer to the first storage element of the slice.
+	/// Two-element bitfield structure, holding pointer and head information.
 	///
-	/// This will always be a pointer to one byte, regardless of the storage
-	/// type of the `BitSlice` or the type parameter of `Self`. It is a
-	/// combination of a correctly typed and aligned pointer to `T`, and the
-	/// index of a byte within that element.
+	/// This stores a pointer to the zeroth element of the slice, and the high
+	/// bits of the head bit cursor. It is typed as a `NonNull<u8>` in order to
+	/// provide null-value optimizations to `Option<BitPtr<T>>`, and because the
+	/// presence of head-bit cursor information in the lowest bits means the
+	/// bit pattern will not uphold alignment properties assumed by
+	/// `NonNull<T>`.
 	///
-	/// It is not necessarily the address of the byte with the first live bit.
-	/// The location of the first live bit within the first element is governed
-	/// by the [`Cursor`] type of the `BitSlice` using this structure.
+	/// This field cannot be treated as an address of the zeroth byte of the
+	/// slice domain, because the owning handle’s [`Cursor`] implementation
+	/// governs the bit pattern of the head cursor.
 	///
 	/// [`Cursor`]: ../trait.Cursor.html
 	ptr: NonNull<u8>,
@@ -208,12 +204,14 @@ where T: Bits {
 	/// The number of high bits in `self.ptr` that are actually the address of
 	/// the zeroth `T`.
 	pub const PTR_DATA_BITS: usize = PTR_BITS - Self::PTR_HEAD_BITS;
+
 	/// Marks the bits of `self.ptr` that are the `data` section.
-	pub const PTR_DATA_MASK: usize = !0 & !Self::PTR_HEAD_MASK;
+	pub const PTR_DATA_MASK: usize = !Self::PTR_HEAD_MASK;
 
 	/// The number of low bits in `self.ptr` that are the high bits of the head
 	/// `BitIdx` cursor.
-	pub const PTR_HEAD_BITS: usize = T::BITS as usize - Self::LEN_HEAD_BITS;
+	pub const PTR_HEAD_BITS: usize = T::INDX as usize - Self::LEN_HEAD_BITS;
+
 	/// Marks the bits of `self.ptr` that are the `head` section.
 	pub const PTR_HEAD_MASK: usize = T::MASK as usize >> Self::LEN_HEAD_BITS;
 
@@ -223,38 +221,44 @@ where T: Bits {
 	/// This is always `3`, until Rust tries to target a machine whose bytes are
 	/// not eight bits wide.
 	pub const LEN_HEAD_BITS: usize = 3;
+
 	/// Marks the bits of `self.len` that are the `head` section.
-	pub const LEN_HEAD_MASK: usize = 7;
+	pub const LEN_HEAD_MASK: usize = 0b0111;
 
 	/// The number of middle bits in `self.len` that are the tail `BitIdx`
 	/// cursor.
-	pub const LEN_TAIL_BITS: usize = T::BITS as usize;
+	pub const LEN_TAIL_BITS: usize = T::INDX as usize;
+
 	/// Marks the bits of `self.len` that are the `tail` section.
 	pub const LEN_TAIL_MASK: usize = (T::MASK as usize) << Self::LEN_HEAD_BITS;
 
 	/// The number of high bits in `self.len` that are used to count `T`
 	/// elements in the slice.
-	pub const LEN_DATA_BITS: usize = USZ_BITS - Self::LEN_INDX_BITS;
+	pub const LEN_ELTS_BITS: usize = USZ_BITS - Self::LEN_INDX_BITS;
+
 	/// Marks the bits of `self.len` that are the `data` section.
-	pub const LEN_DATA_MASK: usize = !0 & !Self::LEN_INDX_MASK;
+	pub const LEN_ELTS_MASK: usize = !Self::LEN_INDX_MASK;
 
 	/// The number of bits occupied by the `tail` `BitIdx` and the low 3 bits of
 	/// `head`.
 	pub const LEN_INDX_BITS: usize = Self::LEN_TAIL_BITS + Self::LEN_HEAD_BITS;
+
 	/// Marks the bits of `self.len` that are either `tail` or `head`.
 	pub const LEN_INDX_MASK: usize = Self::LEN_TAIL_MASK | Self::LEN_HEAD_MASK;
 
-	/// The maximum number of elements that can be stored in a `BitPtr` domain.
-	pub const MAX_ELTS: usize = 1 << Self::LEN_DATA_BITS;
+	/// The inclusive maximum number of elements that can be stored in a
+	/// `BitPtr` domain.
+	pub const MAX_ELTS: usize = 1 << Self::LEN_ELTS_BITS;
 
-	/// The maximum number of bits that can be stored in a `BitPtr` domain.
+	/// The inclusive maximum number of bits that can be stored in a `BitPtr`
+	/// domain.
 	pub const MAX_BITS: usize = !0 >> Self::LEN_HEAD_BITS;
 
 	/// Produces an empty-slice representation.
 	///
 	/// This has no live bits, and has a dangling pointer. It is useful as a
 	/// default value (and is the function used by `Default`) to indicate
-	/// arbitrary null slices.
+	/// arbitrary empty slices.
 	///
 	/// # Returns
 	///
@@ -296,7 +300,9 @@ where T: Bits {
 		//  Null pointers are always well aligned.
 		assert!(
 			(ptr as usize).trailing_zeros() as usize >= Self::PTR_HEAD_BITS,
-			"BitPtr domain pointers must be well aligned",
+			"Pointer {:p} does not satisfy minimum alignment requirements {}",
+			ptr,
+			Self::PTR_HEAD_BITS,
 		);
 		Self {
 			_ty: PhantomData,
@@ -311,26 +317,27 @@ where T: Bits {
 	///
 	/// - `data`: A well-aligned pointer to a storage element. If this is null,
 	///   then the empty-slice representation is returned, regardless of other
-	///   parameter values.
+	///   argument values.
 	/// - `elts`: A number of storage elements in the domain of the new
-	///   `BitPtr`. This number must be in `0 .. Self::MAX_ELTS`.
+	///   `BitPtr`. This number must be in `0 .. Self::MAX_ELTS`. If it is zero,
+	///   then the empty-slice representation is returned, regardless of other
+	///   argument values.
 	/// - `head`: The bit index of the first live bit in the domain. This must
-	///   be in the domain `0 .. T::SIZE`.
+	///   be in the domain `0 .. T::BITS`.
 	/// - `tail`: The bit index of the first dead bit after the domain. This
 	///   must be:
 	///   - equal to `head` when `elts` is `1`, to create an empty slice.
-	///   - in `head + 1 ..= T::SIZE` when `elts` is `1` to create a
+	///   - in `head + 1 ..= T::BITS` when `elts` is `1` to create a
 	///     single-element slice.
-	///   - in `1 ..= T::SIZE` when `elts` is greater than `1`.
-	///   - in `1 .. T::SIZE` when `elts` is `Self::MAX_ELTS - 1`.
+	///   - in `1 ..= T::BITS` when `elts` is greater than `1`.
+	///   - in `1 .. T::BITS` when `elts` is `Self::MAX_ELTS`.
 	///
 	/// # Returns
 	///
-	/// If `data` is null, then the empty slice is returned.
-	///
-	/// If either of the following conditions are true, then the uninhabited
-	/// slice is returned:
+	/// If any of the following conditions are true, then an empty slice is
+	/// returned:
 	///
+	/// - `data` is the null pointer,
 	/// - `elts` is `0`,
 	/// - `elts` is `1` **and** `head` is equal to `tail`.
 	///
@@ -357,16 +364,18 @@ where T: Bits {
 	/// # Safety
 	///
 	/// The `data` pointer and `elts` counter must describe a correctly aligned,
-	/// validly allocated, region of memory. The caller is responsible for
-	/// ensuring that the slice of memory that the new `BitPtr` will govern is
-	/// all governable.
-	pub fn new<Head: Into<BitIdx>, Tail: Into<BitIdx>>(
+	/// validly allocated, region of memory, or the canonical dangling pointer
+	/// and empty slice. The caller is responsible for ensuring that the slice
+	/// of memory that the new `BitPtr` will govern is all governable.
+	pub fn new<Head, Tail>(
 		data: *const T,
 		elts: usize,
 		head: Head,
 		tail: Tail,
-	) -> Self {
+	) -> Self
+	where Head: Into<BitIdx>, Tail: Into<BitIdx> {
 		let (head, tail) = (head.into(), tail.into());
+
 		//  null pointers, and pointers to empty regions, are run through the
 		//  uninhabited constructor instead
 		if data.is_null() || elts == 0 || (elts == 1 && head == tail) {
@@ -381,29 +390,26 @@ where T: Bits {
 
 		//  Check that the slice domain is below the ceiling.
 		assert!(
-			elts < Self::MAX_ELTS,
-			"BitPtr domain regions must have at most {} elements",
-			Self::MAX_ELTS - 1,
+			elts <= Self::MAX_ELTS,
+			"{} exceeds the BitPtr domain maximum, {}",
+			elts,
+			Self::MAX_ELTS,
 		);
 
-		//  Check that the pointer is not so high in the address space that the
-		//  slice domain wraps.
-		if data.wrapping_offset(elts as isize) < data {
-			panic!("BitPtr slices MUST NOT wrap around the address space");
-		}
-
 		//  Check that the head cursor index is within the storage element.
 		assert!(
 			head.is_valid::<T>(),
-			"BitPtr head cursors must be in the domain 0 .. {}",
-			T::SIZE,
+			"{} is outside the head domain 0 .. {}",
+			*head,
+			T::BITS,
 		);
 
 		//  Check that the tail cursor index is in the appropriate domain.
 		assert!(
-			BitIdx::from(*tail - 1).is_valid::<T>(),
-			"BitPtr tail cursors must be in the domain 1 ..= {}",
-			T::SIZE,
+			tail.is_valid_tail::<T>(),
+			"{} is outside the tail domain 1 ..= {}",
+			*tail,
+			T::BITS,
 		);
 
 		//  For single-element slices, check that the tail cursor is after the
@@ -415,21 +421,67 @@ where T: Bits {
 				beyond the head cursor",
 			);
 		}
-		else if elts == Self::MAX_ELTS - 1 {
+		else if elts == Self::MAX_ELTS {
 			assert!(
 				tail.is_valid::<T>(),
-				"BitPtr domains with maximum elements must have the tail \
+				"BitPtr domains with maximum elements must have the tail ({}) \
 				cursor in 1 .. {}",
-				T::SIZE,
+				*tail,
+				T::BITS,
 			);
 		}
 
-		//  All invariants satisfied; build the fields
+		unsafe { Self::new_unchecked(data, elts, head, tail) }
+	}
+
+	/// Creates a new `BitPtr<T>` from its components, without validity checks.
+	///
+	/// # Safety
+	///
+	/// ***NONE.*** This function *only* packs its parameters into the bit
+	/// pattern of the `BitPtr<T>` type. It should only be used in contexts
+	/// where a previously extant `BitPtr<T>` was constructed via `new`, and
+	/// its `raw_parts()` components are known to be valid for reconstruction.
+	///
+	/// This is a bypass for the validity checks, and can only be used when
+	/// the checks are statically known to still be in force.
+	///
+	/// # Parameters
+	///
+	/// Same as [`::new`], except `head` and `tail` are concretely `BitIdx`.
+	///
+	/// # Returns
+	///
+	/// See `::new`.
+	///
+	/// [`::new`]: #method.new
+	pub(crate) unsafe fn new_unchecked(
+		data: *const T,
+		elts: usize,
+		head: BitIdx,
+		tail: BitIdx,
+	) -> Self {
+		//  This check cannot ever be elided. Check that the pointer is not so
+		//  high in the address space that the slice domain wraps.
+		assert!(
+			data.wrapping_add(elts) >= data,
+			"The region overflows the address space: {:p} + {:02X} is {:p}",
+			data,
+			elts,
+			data.wrapping_add(elts),
+		);
+
 		let ptr_data = data as usize & Self::PTR_DATA_MASK;
 		let ptr_head = *head as usize >> Self::LEN_HEAD_BITS;
 
-		let len_elts = elts << Self::LEN_INDX_BITS;
-		//  Store tail. Note that this wraps T::SIZE to 0. This must be
+		//  If `tail` is not maximal, it will be wrapped to zero, and `elts`
+		//  will be incremented. Since `elts` is about to be unconditionally
+		//  decremented, the equivalent behavior is to not increment, and
+		//  decrement only if `tail` is not maximal.
+		let len_elts = elts.saturating_sub((*tail < T::BITS) as usize)
+			<< Self::LEN_INDX_BITS;
+
+		//  Store tail. Note that this wraps `T::BITS` to 0. This must be
 		//  reconstructed during retrieval.
 		let len_tail
 			= ((*tail as usize) << Self::LEN_HEAD_BITS)
@@ -438,9 +490,7 @@ where T: Bits {
 
 		Self {
 			_ty: PhantomData,
-			ptr: unsafe {
-				NonNull::new_unchecked((ptr_data | ptr_head) as *mut u8)
-			},
+			ptr: NonNull::new_unchecked((ptr_data | ptr_head) as *mut u8),
 			len: len_elts | len_tail | len_head,
 		}
 	}
@@ -458,7 +508,7 @@ where T: Bits {
 	/// # Safety
 	///
 	/// This pointer must be valid in the user’s memory model and allocation
-	/// regime.
+	/// regime in order for the caller to dereference it.
 	pub fn pointer(&self) -> *const T {
 		(self.ptr.as_ptr() as usize & Self::PTR_DATA_MASK) as *const T
 	}
@@ -478,7 +528,11 @@ where T: Bits {
 	/// This size must be valid in the user’s memory model and allocation
 	/// regime.
 	pub fn elements(&self) -> usize {
-		self.len >> Self::LEN_INDX_BITS
+		//  Count the elements as marked in the elts field, adding one unless
+		//  the tail is `T::BITS` or `0`.
+		let tail_bits = self.len & Self::LEN_TAIL_MASK;
+		let incr = tail_bits != 0;
+		(self.len >> Self::LEN_INDX_BITS) + incr as usize
 	}
 
 	/// Extracts the element cursor of the head bit.
@@ -490,10 +544,12 @@ where T: Bits {
 	/// # Returns
 	///
 	/// A `BitIdx` that is the index of the first live bit in the first element.
-	/// This will be in the domain `0 .. T::SIZE`.
+	/// This will be in the domain `0 .. T::BITS`.
 	pub fn head(&self) -> BitIdx {
-		((((self.ptr.as_ptr() as usize & Self::PTR_HEAD_MASK) << 3)
-		| (self.len & Self::LEN_HEAD_MASK)) as u8).into()
+		let ptr = self.ptr.as_ptr() as usize & Self::PTR_HEAD_MASK;
+		let ptr = ptr << Self::LEN_HEAD_BITS;
+		let len = self.len & Self::LEN_HEAD_MASK;
+		((ptr | len) as u8).into()
 	}
 
 	/// Extracts the element cursor of the first dead bit *after* the tail bit.
@@ -505,16 +561,38 @@ where T: Bits {
 	/// # Returns
 	///
 	/// A `BitIdx` that is the index of the first dead bit after the last live
-	/// bit in the last element. This will be in the domain `1 ..= T::SIZE`.
+	/// bit in the last element. This will be in the domain `1 ..= T::BITS`.
 	pub fn tail(&self) -> BitIdx {
+		/* This function is one of the most-used in the library. As such, its
+		 * implementation is written in a straight linear style. The compiler is
+		 * free to rearrange the code as it sees fit, and may not reflect the
+		 * code below. The equivalent, user-friendly function body is:
+
+		if self.is_empty() {
+			return 0.into();
+		}
 		let bits = (self.len & Self::LEN_TAIL_MASK) >> Self::LEN_HEAD_BITS;
-		if bits == 0 { T::SIZE } else { bits as u8 }.into()
+		if bits == 0 { T::BITS } else { bits as u8 }.into()
+		 */
+
+		//  The tail is stored in the LEN_TAIL_MASK region.
+		let bits = (self.len & Self::LEN_TAIL_MASK) >> Self::LEN_HEAD_BITS;
+		(
+			//  Becomes 0 when the slice is empty, and !0 when not. This clamps
+			//  the tail to 0 on the empty slice, or itself on non-empty.
+			//  `0 & rhs` is always `0`, `!0 & rhs` is always `rhs`.
+			(self.is_empty() as u8).wrapping_sub(1) &
+			//  If the tail’s bit pattern is zero, wrap it to the maximal. This
+			//  upshifts `1` (pattern is zero) or `0` (pattern is not), then
+			//  sets the upshift bit on the pattern.
+			((((bits == 0) as u8) << T::INDX) | bits as u8)
+		).into()
 	}
 
 	/// Decomposes the pointer into raw components.
 	///
-	/// The values returned from this can be immediately passed into `::new` in
-	/// order to rebuild the pointer.
+	/// The values returned from this can be immediately passed into `::new` or
+	/// `::new_unchecked` in order to rebuild the pointer.
 	///
 	/// # Parameters
 	///
@@ -524,12 +602,42 @@ where T: Bits {
 	///
 	/// - `*const T`: A well aligned pointer to the first element of the slice.
 	/// - `usize`: The number of elements in the slice.
-	/// - `head`: The index of the first live bit in the first element.
-	/// - `tail`: The index of the first dead bit in the last element.
+	/// - `BitIdx`: The index of the first live bit in the first element.
+	/// - `BitIdx`: The index of the first dead bit in the last element.
 	pub fn raw_parts(&self) -> (*const T, usize, BitIdx, BitIdx) {
 		(self.pointer(), self.elements(), self.head(), self.tail())
 	}
 
+	/// Produces the element count, head index, and tail index, which describe
+	/// the region.
+	///
+	/// # Parameters
+	///
+	/// - `&self`
+	///
+	/// # Returns
+	///
+	/// - `usize`: The number of elements in the slice.
+	/// - `BitIdx`: The index of the first live bit in the first element.
+	/// - `BitIdx`: The index of the first dead bit in the last element.
+	pub fn region_data(&self) -> (usize, BitIdx, BitIdx) {
+		(self.elements(), self.head(), self.tail())
+	}
+
+	/// Produces the head and tail indices for the slice.
+	///
+	/// # Parameters
+	///
+	/// - `&self`
+	///
+	/// # Returns
+	///
+	/// - `BitIdx`: The index of the first live bit in the first element.
+	/// - `BitIdx`: The index of the first dead bit in the last element.
+	pub fn cursors(&self) -> (BitIdx, BitIdx) {
+		(self.head(), self.tail())
+	}
+
 	/// Checks if the pointer represents the empty slice.
 	///
 	/// The empty slice has a dangling `data` pointer and zeroed `elts`, `head`,
@@ -543,30 +651,10 @@ where T: Bits {
 	///
 	/// Whether the slice is empty or inhabited.
 	pub fn is_empty(&self) -> bool {
-		self.len >> Self::LEN_INDX_BITS == 0
+		self.len == 0 && self.ptr.as_ptr() as usize & Self::PTR_HEAD_MASK == 0
 	}
 
-	/// Checks if the pointer represents the full slice.
-	///
-	/// The full slice is marked by `0` values for `elts` and `tail`, when
-	/// `data` is not null. The full slice does not need `head` to be `0`.
-	///
-	/// # Parameters
-	///
-	/// - `&self`
-	///
-	/// # Returns
-	///
-	/// Whether the slice is fully extended or not.
-	pub fn is_full(&self) -> bool {
-		//  Self must be:
-		//  - not empty
-		//  - `!0` in `elts` and `tail`
-		!self.is_empty()
-		&& ((self.len | Self::LEN_HEAD_MASK) == !0)
-	}
-
-	///  Counts how many bits are in the domain of a `BitPtr` slice.
+	/// Counts how many bits are in the domain of a `BitPtr` slice.
 	///
 	/// # Parameters
 	///
@@ -575,28 +663,21 @@ where T: Bits {
 	/// # Returns
 	///
 	/// A count of the live bits in the slice.
-	pub fn bits(&self) -> usize {
-		if self.is_empty() {
-			return 0;
+	pub fn len(&self) -> usize {
+		let (elts, head, tail) = self.region_data();
+		match elts {
+			0 => 0,
+			1 => (*tail - *head) as usize,
+			//  The number of bits in a domain is calculated by decrementing
+			//  `elts`, multiplying it by the number of bits per element, then
+			//  subtracting `head` (which is the number of dead bits in the
+			//  front of the first element), and adding `tail` (which is the
+			//  number of live bits in the front of the last element).
+			e => ((e - 1) << T::INDX) + (*tail as usize) - (*head as usize),
 		}
-		let (_, elts, head, tail) = self.raw_parts();
-		if elts == 1 {
-			return *tail as usize - *head as usize;
-		}
-		//  The number of bits in a domain is calculated by decrementing `elts`,
-		//  multiplying it by the number of bits per element, then subtracting
-		//  `head` (which is the number of dead bits in the front of the first
-		//  element), and adding `tail` (which is the number of live bits in the
-		//  front of the last element).
-		((elts - 1) << T::BITS)
-			.saturating_add(*tail as usize)
-			.saturating_sub(*head as usize)
 	}
 
-	/// Produces the head element, if and only if it is partially live.
-	///
-	/// If the head element is completely live, this returns `None`, because the
-	/// head element is returned in `body_elts()`.
+	/// Accesses the element slice behind the pointer as a Rust slice.
 	///
 	/// # Parameters
 	///
@@ -604,22 +685,16 @@ where T: Bits {
 	///
 	/// # Returns
 	///
-	/// `Some(&T)` if the slice has at least one element, and the first element
-	/// has at least one bit dead.
+	/// Standard Rust slice handle over the data governed by this pointer.
 	///
-	/// `None` if the slice is empty, or if the first element is completely
-	/// live.
-	pub fn head_elt(&self) -> Option<&T> {
-		if !self.is_empty() && *self.head() > 0 {
-			return Some(&self.as_ref()[0]);
-		}
-		None
+	/// # Lifetimes
+	///
+	/// - `'a`: Lifetime for which the data behind the pointer is live.
+	pub fn as_slice<'a>(&self) -> &'a [T] {
+		unsafe { slice::from_raw_parts(self.pointer(), self.elements()) }
 	}
 
-	/// Produces the slice of middle elements that are all fully live.
-	///
-	/// This may produce the empty slice, if the `BitPtr` slice domain has zero,
-	/// one, or two elements, and the outer elements are only partially live.
+	/// Accesses the element slice behind the pointer as a Rust mutable slice.
 	///
 	/// # Parameters
 	///
@@ -627,32 +702,18 @@ where T: Bits {
 	///
 	/// # Returns
 	///
-	/// A slice of fully live storage elements.
-	pub fn body_elts(&self) -> &[T] {
-		let w: u8 = 1 << Self::LEN_TAIL_BITS;
-		let (_, e, h, t) = self.raw_parts();
-		match (e, *h, *t) {
-			//  Empty slice
-			(0, _, _)           => &             [          ],
-			//  Single-element slice, with cursors at the far edges
-			(1, 0, t) if t == w => &self.as_ref()[0 .. e - 0],
-			//  Single-element slice, with partial cursors
-			(1, _, _)           => &             [          ],
-			//  Multiple-element slice, with cursors at the far edges
-			(_, 0, t) if t == w => &self.as_ref()[0 .. e - 0],
-			//  Multiple-element slice, with full head and partial tail
-			(_, 0, _)           => &self.as_ref()[0 .. e - 1],
-			//  Multiple-element slice, with partial tail and full head
-			(_, _, t) if t == w => &self.as_ref()[1 .. e - 0],
-			//  Multiple-element slice, with partial cursors
-			(_, _, _)           => &self.as_ref()[1 .. e - 1],
+	/// Standard Rust slice handle over the data governed by this pointer.
+	///
+	/// # Lifetimes
+	///
+	/// - `'a`: Lifetime for which the data behind the pointer is live.
+	pub fn as_mut_slice<'a>(&self) -> &'a mut [T] {
+		unsafe {
+			slice::from_raw_parts_mut(self.pointer() as *mut T, self.elements())
 		}
 	}
 
-	/// Produces the tail element, if and only if it is partially live.
-	///
-	/// If the tail element is completely live, this returns `None`, because the
-	/// tail element is returned in `body_elts()`.
+	/// Gets the domain kind for the region the pointer describes.
 	///
 	/// # Parameters
 	///
@@ -660,43 +721,37 @@ where T: Bits {
 	///
 	/// # Returns
 	///
-	/// `Some(&T)` if the slice has at least one element, and the last element
-	/// has at least one bit dead.
-	///
-	/// `None` if the slice is empty, or if the last element is completely live.
-	pub fn tail_elt(&self) -> Option<&T> {
-		if !self.is_empty() && *self.tail() < T::SIZE {
-			return Some(&self.as_ref()[self.elements() - 1]);
-		}
-		None
+	/// An enum describing the live bits in the region the pointer covers.
+	pub fn domain_kind(&self) -> BitDomainKind {
+		self.into()
 	}
 
-	pub fn set_head<Head: Into<BitIdx>>(&mut self, head: Head) {
-		if self.is_empty() {
-			return;
-		}
-		let head = head.into();
-		assert!(
-			head.is_valid::<T>(),
-			"Head indices must be in the domain 0 .. {}",
-			T::SIZE,
-		);
-		if self.elements() == 1 {
-			assert!(
-				head <= self.tail(),
-				"Single-element slices must have head below tail",
-			);
-		}
-		self.ptr = unsafe {
-			let ptr = self.ptr.as_ptr() as usize;
-			NonNull::new_unchecked(
-				((ptr & !Self::PTR_HEAD_MASK)
-				| ((*head as usize >> Self::LEN_HEAD_BITS) & Self::PTR_HEAD_MASK)
-				) as *mut u8
-			)
-		};
-		self.len &= !Self::LEN_HEAD_MASK;
-		self.len |= *head as usize & Self::LEN_HEAD_MASK;
+	/// Gets the domain for the region the pointer describes.
+	///
+	/// # Parameters
+	///
+	/// - `&self`
+	///
+	/// # Returns
+	///
+	/// An enum containing the logical components of the domain governed by
+	/// `self`.
+	pub fn domain(&self) -> BitDomain<T> {
+		(*self).into()
+	}
+
+	/// Gets the domain for the region the pointer describes.
+	///
+	/// # Parameters
+	///
+	/// - `&self`
+	///
+	/// # Returns
+	///
+	/// An enum containing the logical components of the domain governed by
+	/// `self`.
+	pub fn domain_mut(&self) -> BitDomainMut<T> {
+		(*self).into()
 	}
 
 	/// Moves the `head` cursor upwards by one.
@@ -716,13 +771,26 @@ where T: Bits {
 	/// solely responsible for owned memory, its conception of the allocation
 	/// will differ from the allocator’s.
 	pub unsafe fn incr_head(&mut self) {
-		let (data, elts, head, tail) = self.raw_parts();
-		let (new_head, wrap) = head.incr::<T>();
-		if wrap {
-			*self = Self::new(data.offset(1), elts - 1, new_head, tail);
-		}
-		else {
-			*self = Self::new(data, elts, new_head, tail);
+		match self.len() {
+			//  Do nothing when empty
+			0 => {},
+			//  Become the uninhabited slice when the last bit is removed
+			1 => *self = Self::uninhabited(self.pointer()),
+			_ => {
+				let (data, elts, head, tail) = self.raw_parts();
+				let (h, wrap) = head.incr::<T>();
+				if wrap {
+					*self = Self::new_unchecked(
+						data.offset(1),
+						elts.saturating_sub(1),
+						h,
+						tail,
+					);
+				}
+				else {
+					*self = Self::new_unchecked(data, elts, h, tail);
+				}
+			},
 		}
 	}
 
@@ -743,34 +811,23 @@ where T: Bits {
 	/// solely responsible for owned memory, its conception of the allocation
 	/// will differ from the allocator’s.
 	pub unsafe fn decr_head(&mut self) {
-		let (data, elts, head, tail) = self.raw_parts();
-		let (new_head, wrap) = head.decr::<T>();
-		if wrap {
-			*self = Self::new(data.offset(-1), elts + 1, new_head, tail);
-		}
-		else {
-			*self = Self::new(data, elts, new_head, tail);
-		}
-	}
-
-	pub fn set_tail<Tail: Into<BitIdx>>(&mut self, tail: Tail) {
+		//  Empty slices cannot be modified in this way
 		if self.is_empty() {
 			return;
 		}
-		let tail = tail.into();
-		assert!(
-			BitIdx::from(*tail - 1).is_valid::<T>(),
-			"Tail indices must be in the domain 1 ..= {}",
-			T::SIZE,
-		);
-		if self.elements() == 1 {
-			assert!(
-				tail >= self.head(),
-				"Single-element slices must have tail above head",
+		let (data, elts, head, tail) = self.raw_parts();
+		let (h, wrap) = head.decr::<T>();
+		if wrap {
+			*self = Self::new_unchecked(
+				data.offset(-1),
+				elts.saturating_add(1),
+				h,
+				tail,
 			);
 		}
-		self.len &= !Self::LEN_TAIL_MASK;
-		self.len |= *tail as usize
+		else {
+			*self = Self::new_unchecked(data, elts, h, tail);
+		}
 	}
 
 	/// Moves the `tail` cursor upwards by one.
@@ -790,11 +847,18 @@ where T: Bits {
 	/// responsible for owned memory, its conception of the allocation will
 	/// differ from the allocator’s.
 	pub unsafe fn incr_tail(&mut self) {
+		//  Fully extended slices cannot have their tail increased.
+		if self.len | Self::LEN_HEAD_MASK == !0 {
+			return;
+		}
 		let (data, elts, head, tail) = self.raw_parts();
-		let decr = BitIdx::from(*tail - 1);
-		let (mut new_tail, wrap) = decr.incr::<T>();
-		new_tail = BitIdx::from(*new_tail + 1);
-		*self = Self::new(data, elts + wrap as usize, head, new_tail);
+		let (t, wrap) = tail.incr_tail::<T>();
+		*self = Self::new_unchecked(
+			data,
+			elts.saturating_add(wrap as usize),
+			head,
+			t,
+		);
 	}
 
 	/// Moves the `tail` cursor downwards by one.
@@ -814,11 +878,80 @@ where T: Bits {
 	/// responsible for owned memory, its conception of the allocation will
 	/// differ from the allocator’s.
 	pub unsafe fn decr_tail(&mut self) {
-		let (data, elts, head, tail) = self.raw_parts();
-		let decr = BitIdx::from(*tail - 1);
-		let (mut new_tail, wrap) = decr.decr::<T>();
-		new_tail = BitIdx::from(*new_tail + 1);
-		*self = Self::new(data, elts - wrap as usize, head, new_tail);
+		match self.len() {
+			//  Empty slices cannot have their tail decremented
+			0 => return,
+			//  One-bit slices become uninhabited
+			1 => *self = Self::uninhabited(self.pointer()),
+			_ => {
+				let (data, elts, head, tail) = self.raw_parts();
+				let (t, wrap) = tail.decr_tail::<T>();
+				*self = Self::new_unchecked(
+					data,
+					elts.saturating_sub(wrap as usize),
+					head,
+					t,
+				);
+			},
+		}
+	}
+
+	/// Converts a `BitSlice` handle into its `BitPtr` representation.
+	///
+	/// # Parameters
+	///
+	/// - `bs: &BitSlice<C, T>`: a `BitSlice` handle
+	///
+	/// # Returns
+	///
+	/// The `BitPtr<T>` structure composing the handle.
+	pub(crate) fn from_bitslice<C>(bs: &BitSlice<C, T>) -> Self
+	where C: Cursor {
+		let src = unsafe { &*(bs as *const BitSlice<C, T> as *const [()]) };
+		let (ptr, len) = match (src.as_ptr(), src.len()) {
+			(_, 0) => (NonNull::dangling(), 0),
+			(p, _) if p.is_null() => unreachable!("Rust forbids null refs"),
+			(p, l) => (unsafe { NonNull::new_unchecked(p as *mut u8) }, l),
+		};
+		Self { ptr, len, _ty: PhantomData }
+	}
+
+	/// Converts a `BitPtr` structure into an immutable `BitSlice` handle.
+	///
+	/// # Parameters
+	///
+	/// - `self`
+	///
+	/// # Returns
+	///
+	/// A `BitSlice` handle composed of the `BitPtr` structure.
+	pub(crate) fn into_bitslice<'a, C>(self) -> &'a BitSlice<C, T>
+	where C: Cursor {
+		unsafe {
+			&*(slice::from_raw_parts(
+				self.ptr.as_ptr() as *const (),
+				self.len,
+			) as *const [()] as *const BitSlice<C, T>)
+		}
+	}
+
+	/// Converts a `BitPtr` structure into a mutable `BitSlice` handle.
+	///
+	/// # Parameters
+	///
+	/// - `self`
+	///
+	/// # Returns
+	///
+	/// A `BitSlice` handle composed of the `BitPtr` structure.
+	pub(crate) fn into_bitslice_mut<'a, C>(self) -> &'a mut BitSlice<C, T>
+	where C: Cursor {
+		unsafe {
+			&mut *(slice::from_raw_parts_mut(
+				self.ptr.as_ptr() as *const () as *mut (),
+				self.len,
+			) as *mut [()] as *mut BitSlice<C, T>)
+		}
 	}
 }
 
@@ -833,9 +966,7 @@ where T: Bits {
 impl<T> AsMut<[T]> for BitPtr<T>
 where T: Bits {
 	fn as_mut(&mut self) -> &mut [T] {
-		let ptr = self.pointer() as *mut T;
-		let len = self.elements();
-		unsafe { slice::from_raw_parts_mut(ptr, len) }
+		self.as_mut_slice()
 	}
 }
 
@@ -844,43 +975,21 @@ where T: Bits {
 impl<T> AsRef<[T]> for BitPtr<T>
 where T: Bits {
 	fn as_ref(&self) -> &[T] {
-		unsafe { slice::from_raw_parts(self.pointer(), self.elements()) }
+		self.as_slice()
 	}
 }
 
-/// Constructs from an immutable `BitSlice` reference handle.
 impl<'a, C, T> From<&'a BitSlice<C, T>> for BitPtr<T>
 where C: Cursor, T: 'a + Bits {
 	fn from(src: &'a BitSlice<C, T>) -> Self {
-		let src: &[()] = unsafe {
-			mem::transmute::<&'a BitSlice<C, T>, &[()]>(src)
-		};
-		let (ptr, len) = match (src.as_ptr() as usize, src.len()) {
-			(_, 0) => (NonNull::dangling(), 0),
-			(0, _) => unreachable!(
-				"Slices cannot have a length when they begin at address 0"
-			),
-			(p, l) => (unsafe { NonNull::new_unchecked(p as *mut u8) }, l),
-		};
-		Self { ptr, len, _ty: PhantomData }
+		Self::from_bitslice(src)
 	}
 }
 
-/// Constructs from a mutable `BitSlice` reference handle.
 impl<'a, C, T> From<&'a mut BitSlice<C, T>> for BitPtr<T>
 where C: Cursor, T: 'a + Bits {
 	fn from(src: &'a mut BitSlice<C, T>) -> Self {
-		let src: &[()] = unsafe {
-			mem::transmute::<&'a mut BitSlice<C, T>, &[()]>(src)
-		};
-		let (ptr, len) = match (src.as_ptr() as usize, src.len()) {
-			(_, 0) => (NonNull::dangling(), 0),
-			(0, _) => unreachable!(
-				"Slices cannot have a length when they begin at address 0"
-			),
-			(p, l) => (unsafe { NonNull::new_unchecked(p as *mut u8) }, l),
-		};
-		Self { ptr, len, _ty: PhantomData }
+		Self::from_bitslice(src)
 	}
 }
 
@@ -916,7 +1025,7 @@ where T: Bits {
 		struct BinAddr<T: Bits>(BitIdx, PhantomData<T>);
 		impl<T: Bits>  Debug for BinAddr<T> {
 			fn fmt(&self, f: &mut Formatter) -> fmt::Result {
-				f.write_fmt(format_args!("0b{:0>1$b}", *self.0, T::BITS as usize))
+				f.write_fmt(format_args!("0b{:0>1$b}", *self.0, T::INDX as usize))
 			}
 		}
 		write!(f, "BitPtr<{}>", T::TYPENAME)?;
@@ -937,12 +1046,12 @@ mod tests {
 	fn associated_consts_u8() {
 		assert_eq!(BitPtr::<u8>::PTR_DATA_BITS, PTR_BITS);
 		assert_eq!(BitPtr::<u8>::PTR_HEAD_BITS, 0);
-		assert_eq!(BitPtr::<u8>::LEN_DATA_BITS, USZ_BITS - 6);
+		assert_eq!(BitPtr::<u8>::LEN_ELTS_BITS, USZ_BITS - 6);
 		assert_eq!(BitPtr::<u8>::LEN_TAIL_BITS, 3);
 
 		assert_eq!(BitPtr::<u8>::PTR_DATA_MASK, !0);
 		assert_eq!(BitPtr::<u8>::PTR_HEAD_MASK, 0);
-		assert_eq!(BitPtr::<u8>::LEN_DATA_MASK, !0 << 6);
+		assert_eq!(BitPtr::<u8>::LEN_ELTS_MASK, !0 << 6);
 		assert_eq!(BitPtr::<u8>::LEN_TAIL_MASK, 7 << 3);
 		assert_eq!(BitPtr::<u8>::LEN_INDX_MASK, 63);
 	}
@@ -951,12 +1060,12 @@ mod tests {
 	fn associated_consts_u16() {
 		assert_eq!(BitPtr::<u16>::PTR_DATA_BITS, PTR_BITS - 1);
 		assert_eq!(BitPtr::<u16>::PTR_HEAD_BITS, 1);
-		assert_eq!(BitPtr::<u16>::LEN_DATA_BITS, USZ_BITS - 7);
+		assert_eq!(BitPtr::<u16>::LEN_ELTS_BITS, USZ_BITS - 7);
 		assert_eq!(BitPtr::<u16>::LEN_TAIL_BITS, 4);
 
 		assert_eq!(BitPtr::<u16>::PTR_DATA_MASK, !0 << 1);
 		assert_eq!(BitPtr::<u16>::PTR_HEAD_MASK, 1);
-		assert_eq!(BitPtr::<u16>::LEN_DATA_MASK, !0 << 7);
+		assert_eq!(BitPtr::<u16>::LEN_ELTS_MASK, !0 << 7);
 		assert_eq!(BitPtr::<u16>::LEN_TAIL_MASK, 15 << 3);
 		assert_eq!(BitPtr::<u16>::LEN_INDX_MASK, 127);
 	}
@@ -965,12 +1074,12 @@ mod tests {
 	fn associated_consts_u32() {
 		assert_eq!(BitPtr::<u32>::PTR_DATA_BITS, PTR_BITS - 2);
 		assert_eq!(BitPtr::<u32>::PTR_HEAD_BITS, 2);
-		assert_eq!(BitPtr::<u32>::LEN_DATA_BITS, USZ_BITS - 8);
+		assert_eq!(BitPtr::<u32>::LEN_ELTS_BITS, USZ_BITS - 8);
 		assert_eq!(BitPtr::<u32>::LEN_TAIL_BITS, 5);
 
 		assert_eq!(BitPtr::<u32>::PTR_DATA_MASK, !0 << 2);
 		assert_eq!(BitPtr::<u32>::PTR_HEAD_MASK, 3);
-		assert_eq!(BitPtr::<u32>::LEN_DATA_MASK, !0 << 8);
+		assert_eq!(BitPtr::<u32>::LEN_ELTS_MASK, !0 << 8);
 		assert_eq!(BitPtr::<u32>::LEN_TAIL_MASK, 31 << 3);
 		assert_eq!(BitPtr::<u32>::LEN_INDX_MASK, 255);
 	}
@@ -979,12 +1088,12 @@ mod tests {
 	fn associated_consts_u64() {
 		assert_eq!(BitPtr::<u64>::PTR_DATA_BITS, PTR_BITS - 3);
 		assert_eq!(BitPtr::<u64>::PTR_HEAD_BITS, 3);
-		assert_eq!(BitPtr::<u64>::LEN_DATA_BITS, USZ_BITS - 9);
+		assert_eq!(BitPtr::<u64>::LEN_ELTS_BITS, USZ_BITS - 9);
 		assert_eq!(BitPtr::<u64>::LEN_TAIL_BITS, 6);
 
 		assert_eq!(BitPtr::<u64>::PTR_DATA_MASK, !0 << 3);
 		assert_eq!(BitPtr::<u64>::PTR_HEAD_MASK, 7);
-		assert_eq!(BitPtr::<u64>::LEN_DATA_MASK, !0 << 9);
+		assert_eq!(BitPtr::<u64>::LEN_ELTS_MASK, !0 << 9);
 		assert_eq!(BitPtr::<u64>::LEN_TAIL_MASK, 63 << 3);
 		assert_eq!(BitPtr::<u64>::LEN_INDX_MASK, 511);
 	}
@@ -1007,25 +1116,36 @@ mod tests {
 	}
 
 	#[test]
-	fn full() {
-		let elt_ct = BitPtr::<u64>::MAX_ELTS - 1;
-		//  maximum elements, maximum bits
-		let bp = BitPtr::<u64>::new(8 as *const u64, elt_ct, 0, 63);
-		assert!(bp.is_full());
-
-		//  one bit fewer
-		let bp = BitPtr::<u64>::new(8 as *const u64, elt_ct, 0, 62);
-		assert!(!bp.is_full());
-		assert_eq!(*bp.tail(), 62.into());
-
-		//  one element fewer
-		let bp = BitPtr::<u64>::new(8 as *const u64, elt_ct - 1, 0, 64);
-		assert!(!bp.is_full());
+	#[should_panic]
+	fn overfull() {
+		BitPtr::<u32>::new(8 as *const u32, BitPtr::<u32>::MAX_ELTS, 0, 32);
 	}
 
 	#[test]
-	#[should_panic]
-	fn overfull() {
-		BitPtr::<u64>::new(8 as *const u64, BitPtr::<u64>::MAX_ELTS - 1, 0, 64);
+	fn patterns() {
+		fn bp(ptr: *const u32, len: usize) -> BitPtr<u32> {
+			unsafe { core::mem::transmute((ptr, len)) }
+		}
+
+		let ptr = NonNull::<u32>::dangling().as_ptr();
+
+		let h0t0e0 = bp(ptr, 0);
+		assert!(h0t0e0.is_empty());
+
+		let h0t1 = bp(ptr, 8);
+		assert_eq!(*h0t1.head(), 0);
+		assert_eq!(*h0t1.tail(), 1);
+		assert_eq!(h0t1.elements(), 1);
+		assert_eq!(h0t1.len(), 1);
+
+		let h16t17 = bp((ptr as usize | 2) as *const u32, 17 << 3);
+		assert_eq!(*h16t17.head(), 16);
+		assert_eq!(*h16t17.tail(), 17);
+		assert_eq!(h16t17.elements(), 1);
+		assert_eq!(h16t17.len(), 1);
+
+		let h0t0e1 = bp(ptr, 1 << 8);
+		assert_eq!(h0t0e1.elements(), 1);
+		assert_eq!(h0t0e1.len(), 32);
 	}
 }
